# type: ignore

from typing import ClassVar, Optional
from uuid import uuid4

import pandas as pd
import pytest
from pydantic import Field, ValidationError

from neontology.basenode import BaseNode
from neontology.baserelationship import BaseRelationship


class PracticeNode(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "PracticeNode"
    pp: str


class PracticeRelationship(BaseRelationship):
    __relationshiptype__: ClassVar[Optional[str]] = "PRACTICE_RELATIONSHIP"

    source: PracticeNode
    target: PracticeNode

    practice_rel_prop: str = "Default Practice Relationship Property"


def test_base_relationship():
    source_node = PracticeNode(pp="Source Node")

    target_node = PracticeNode(pp="Target Node")

    br = PracticeRelationship(source=source_node, target=target_node)

    assert br.source.pp == "Source Node"
    assert br.target.pp == "Target Node"

    assert br.get_relationship_type() == "PRACTICE_RELATIONSHIP"


def test_rel_schema():
    schema = PracticeRelationship.neontology_schema()

    assert schema.relationship_type == "PRACTICE_RELATIONSHIP"
    assert schema.source_labels == ["PracticeNode"]

    practice_rel_prop = [x for x in schema.properties if x.name == "practice_rel_prop"][
        0
    ]

    assert practice_rel_prop.type_annotation.representation == "str"


def test_source_target_type():
    source_node = "Not a BaseNode"
    target_node = "Also not a BaseNode"

    with pytest.raises(ValidationError) as exception_info:
        PracticeRelationship(source=source_node, target=target_node)

    expected = "2 validation errors for PracticeRelationship"

    assert expected in str(exception_info.value)


def test_merge_relationship(use_graph):
    source_node = PracticeNode(pp="Source Node")
    source_node.create()

    target_node = PracticeNode(pp="Target Node")
    target_node.create()

    br = PracticeRelationship(
        source=source_node,
        target=target_node,
        practice_rel_prop="Default Practice Relationship Property",
    )
    br.merge()

    cypher = """
    MATCH (src:PracticeNode {pp: 'Source Node'})-[r]->(tgt:PracticeNode {pp: 'Target Node'})
    RETURN r.practice_rel_prop
    """

    result = use_graph.evaluate_query_single(cypher)

    assert result == "Default Practice Relationship Property"


def test_match_relationship(use_graph):
    source_node = PracticeNode(pp="Source Node")
    source_node.create()

    target_node = PracticeNode(pp="Target Node")
    target_node.create()

    br = PracticeRelationship(
        source=source_node,
        target=target_node,
        practice_rel_prop="TESTING MATCH RELATIONSHIP",
    )
    br.merge()

    rels = PracticeRelationship.match_relationships()

    assert rels[0].practice_rel_prop == "TESTING MATCH RELATIONSHIP"


class RelMergeOnMatchTest(PracticeRelationship):
    __relationshiptype__: ClassVar[Optional[str]] = "TEST_REL_MERGE_ON_MATCH"
    prop_to_merge_on: str = Field(json_schema_extra={"merge_on": True})
    my_prop: str


def test_merge_relationship_merge_on_match(use_graph):
    source_node = PracticeNode(pp="Source Node")
    source_node.create()

    target_node = PracticeNode(pp="Target Node")
    target_node.create()

    br = RelMergeOnMatchTest(
        source=source_node,
        target=target_node,
        prop_to_merge_on="MergeMe",
        my_prop="Foo",
    )
    br.merge()

    cypher = """
    MATCH (src:PracticeNode {pp: 'Source Node'})-[r:TEST_REL_MERGE_ON_MATCH]->(tgt:PracticeNode {pp: 'Target Node'})
    RETURN r.my_prop
    """

    result = use_graph.evaluate_query_single(cypher)

    assert result == "Foo"

    br2 = RelMergeOnMatchTest(
        source=source_node,
        target=target_node,
        prop_to_merge_on="MergeMe",
        my_prop="Bar",
    )
    br2.merge()

    result2 = use_graph.evaluate_query_single(cypher)

    assert result2 == "Bar"


class RelMergeOnCreateTest(PracticeRelationship):
    __relationshiptype__: ClassVar[Optional[str]] = "TEST_REL_MERGE_ON_CREATE"
    prop_to_merge_on: str = Field(json_schema_extra={"merge_on": True})
    my_prop: str


def test_merge_relationship_merge_on_create(use_graph):
    source_node = PracticeNode(pp="Source Node")
    source_node.create()

    target_node = PracticeNode(pp="Target Node")
    target_node.create()

    br = RelMergeOnCreateTest(
        source=source_node,
        target=target_node,
        prop_to_merge_on="MergeMe",
        my_prop="Foo",
    )
    br.merge()

    cypher = """
    MATCH (src:PracticeNode {pp: 'Source Node'})-[r:TEST_REL_MERGE_ON_CREATE]->(tgt:PracticeNode {pp: 'Target Node'})
    RETURN COLLECT(r.my_prop)
    """

    result = use_graph.evaluate_query_single(cypher)

    assert result == ["Foo"]

    br2 = RelMergeOnCreateTest(
        source=source_node,
        target=target_node,
        prop_to_merge_on="Don'tMergeMe",
        my_prop="Bar",
    )
    br2.merge()

    result2 = use_graph.evaluate_query_single(cypher)

    assert set(result2) == {"Foo", "Bar"}


def test_default_relationship_type():
    source_node = PracticeNode(pp="Source Node")

    target_node = PracticeNode(pp="Target Node")

    br = PracticeRelationship(source=source_node, target=target_node)

    assert br.get_relationship_type() == "PRACTICE_RELATIONSHIP"


def test_default_relationship_type_inherited():
    source_node = PracticeNode(pp="Source Node")

    target_node = PracticeNode(pp="Target Node")

    class NewRelType(BaseRelationship):
        source: PracticeNode
        target: PracticeNode
        __relationshiptype__: ClassVar[Optional[str]] = "NEWRELTYPE"

    rel = NewRelType(source=source_node, target=target_node)

    assert rel.get_relationship_type() == "NEWRELTYPE"


def test_defined_relationship_type_inherited():
    source_node = PracticeNode(pp="Source Node")

    target_node = PracticeNode(pp="Target Node")

    class NewRelType(BaseRelationship):
        source: PracticeNode
        target: PracticeNode
        __relationshiptype__: ClassVar[Optional[str]] = "TEST_RELATIONSHIP_TYPE"

    rel = NewRelType(source=source_node, target=target_node)

    assert rel.get_relationship_type() == "TEST_RELATIONSHIP_TYPE"


class NewRelType(BaseRelationship):
    source: PracticeNode
    target: PracticeNode
    __relationshiptype__: ClassVar[Optional[str]] = "TEST_NEW_RELATIONSHIP_TYPE"

    new_rel_prop: str


def test_merge_relationships_defined_types(use_graph):
    node1 = PracticeNode(pp="Source Node")
    node1.create()

    node2 = PracticeNode(pp="Target Node")
    node2.create()

    rel1 = NewRelType(source=node1, target=node2, new_rel_prop="Rel 1")

    rel2 = NewRelType(source=node2, target=node1, new_rel_prop="Rel 2")

    NewRelType.merge_relationships([rel1, rel2])

    cypher = """
    MATCH (src:PracticeNode)-[r:TEST_NEW_RELATIONSHIP_TYPE]->(tgt:PracticeNode)
    RETURN COLLECT(r.new_rel_prop)
    """

    results = use_graph.evaluate_query_single(cypher)

    assert len(results) == 2

    assert set(results) == {"Rel 1", "Rel 2"}


def test_get_count(use_graph):
    node1 = PracticeNode(pp="Source Node")
    node1.create()

    node2 = PracticeNode(pp="Target Node")
    node2.create()

    rel1 = NewRelType(source=node1, target=node2, new_rel_prop="Rel 1")
    rel2 = NewRelType(source=node2, target=node1, new_rel_prop="Rel 2")

    NewRelType.merge_relationships([rel1, rel2])

    assert NewRelType.get_count() == 2


def test_get_count_none(use_graph):
    assert NewRelType.get_count() == 0


class SubclassNode(PracticeNode):
    __primarylabel__: ClassVar[Optional[str]] = "SubclassNode"
    myprop: str


class NewRelType2(BaseRelationship):
    source: SubclassNode
    target: PracticeNode
    __relationshiptype__: ClassVar[Optional[str]] = "TEST_NEW_RELATIONSHIP_TYPE2"

    new_rel_prop: str


def test_merge_df(use_graph):
    source_node = SubclassNode(pp="Source Node", myprop="Some Value")
    source_node.merge()

    target_node = PracticeNode(pp="Target Node")
    target_node.merge()

    rel_records = [
        {"source": "Source Node", "target": "Target Node", "new_rel_prop": "New Rel 3"}
    ]

    df = pd.DataFrame.from_records(rel_records)

    NewRelType2.merge_df(df, SubclassNode, PracticeNode)

    cypher = """
    MATCH (src:SubclassNode {pp: 'Source Node'})-[r:TEST_NEW_RELATIONSHIP_TYPE2]->(tgt:PracticeNode {pp: 'Target Node'})
    RETURN r.new_rel_prop
    """

    result = use_graph.evaluate_query_single(cypher)

    assert result == "New Rel 3"


def test_merge_df_alt_prop(use_graph):
    source_node = SubclassNode(pp="Source Node", myprop="My Prop Value")
    source_node.merge()

    target_node = PracticeNode(pp="Target Node")
    target_node.merge()

    rel_records = [
        {
            "source": "My Prop Value",
            "target": "Target Node",
            "new_rel_prop": "New Rel 4",
        }
    ]

    df = pd.DataFrame.from_records(rel_records)

    NewRelType2.merge_df(df, SubclassNode, PracticeNode, source_prop="myprop")

    cypher = """
    MATCH (src:SubclassNode {pp: 'Source Node'})-[r:TEST_NEW_RELATIONSHIP_TYPE2]->(tgt:PracticeNode {pp: 'Target Node'})
    RETURN r.new_rel_prop
    """

    result = use_graph.evaluate_query_single(cypher)

    assert result == "New Rel 4"


def OLD_test_merge_df_bad_node_type(use_graph):
    class DifferentNode(PracticeNode):
        __primarylabel__: ClassVar[Optional[str]] = "DifferentNode"
        __primaryproperty__: ClassVar[str] = "myprop"
        myprop: str

    source_node = PracticeNode(pp="something else")

    target_node = PracticeNode(pp="Target Node")

    rels = [NewRelType(source=source_node, target=target_node, new_rel_prop="BAD")]

    with pytest.raises(TypeError):
        NewRelType.merge_relationships(rels, PracticeNode, DifferentNode)


def OLD_test_merge_df_bad_node_type2(use_graph):
    class DifferentNode(PracticeNode):
        __primarylabel__: ClassVar[Optional[str]] = "DifferentNode"
        __primaryproperty__: ClassVar[str] = "myprop"
        myprop: str

    source_node = DifferentNode(myprop="something", pp="something else")

    target_node = PracticeNode(pp="Target Node")

    rels = [NewRelType(source=source_node, target=target_node, new_rel_prop="BAD2")]

    with pytest.raises(TypeError):
        NewRelType.merge_relationships(rels, PracticeNode, PracticeNode)


def test_merge_empty_df():
    df = pd.DataFrame()

    result = PracticeRelationship.merge_df(df)

    assert result is None


def test_merge_records(use_graph):
    source_node = SubclassNode(pp="Source Node", myprop="My Prop Value")
    source_node.merge()

    target_node = PracticeNode(pp="Target Node")
    target_node.merge()

    records = [
        {"source": "Source Node", "target": "Target Node", "new_rel_prop": "New Rel 5"}
    ]

    NewRelType2.merge_records(records)

    cypher = """
    MATCH (src:SubclassNode {pp: 'Source Node'})-[r:TEST_NEW_RELATIONSHIP_TYPE2]->(tgt:PracticeNode {pp: 'Target Node'})
    RETURN r.new_rel_prop
    """

    result = use_graph.evaluate_query_single(cypher)

    assert result == "New Rel 5"


def test_create_mass_rels(use_graph, benchmark):
    practice_records = [{"pp": uuid4().hex} for x in range(1000)]

    records_df = pd.DataFrame.from_records(practice_records)

    PracticeNode.merge_df(records_df)

    assert PracticeNode.get_count() == 1000

    people_rels = [
        {"source": practice_records[x]["pp"], "target": practice_records[x + 1]["pp"]}
        for x in range(999)
    ]

    rels_df = pd.DataFrame.from_records(people_rels)

    benchmark(PracticeRelationship.merge_df, rels_df)

    result = use_graph.evaluate_query_single(
        "MATCH (n)-[r]->(o) RETURN COUNT(DISTINCT(r))"
    )

    assert result == 999



from typing import Any, ClassVar, Optional

from neo4j import GraphDatabase
from neo4j import Result as Neo4jResult

from ..result import NeontologyResult
from .graphengine import GraphEngineBase, GraphEngineConfig
from .neo4jengine import neo4j_records_to_neontology_records


class MemgraphEngine(GraphEngineBase):
    """Graph engine for a Memgraph database.

    Note. the memgraph engine is very similar to the Neo4j engine as they
        can both use the same Neo4j driver
    """

    def __init__(self, config: "MemgraphConfig") -> None:
        """Initialise connection to the Memgraph database.

        Args:
            config (MemgraphConfig): Takes a MemgraphConfig object.
        """
        self.driver = GraphDatabase.driver(
            config.connection_uri,
            auth=(config.connection_username, config.connection_password),
        )

    def verify_connection(self) -> bool:
        """Verify the connection to the Memgraph database.

        Returns:
            bool: True if the connection is successful, False otherwise.
        """
        try:
            self.driver.verify_connectivity()
            return True
        except:  # noqa: E722
            return False

    def close_connection(self) -> None:
        """Close the connection to the Memgraph database."""
        try:
            self.driver.close()
        except AttributeError:
            pass

    def evaluate_query(
        self,
        cypher: str,
        params: dict = {},
        node_classes: dict = {},
        relationship_classes: dict = {},
    ) -> NeontologyResult:
        """Evaluate a Cypher query and return the results as Neontology records.

        Args:
            cypher (str): query to evaluate.
            params (dict, optional): parameters to pass through. Defaults to {}.
            node_classes (dict, optional): mapping of labels to node classes used for populating with results. Defaults to {}.
            relationship_classes (dict, optional): mapping of relationship types to classes used for populating with results.
                Defaults to {}.

        Returns:
            NeontologyResult: Result object containing the records, nodes, relationships, and paths.
        """
        result = self.driver.execute_query(cypher, parameters_=params)

        neo4j_records = result.records
        neontology_records, nodes, rels, paths = neo4j_records_to_neontology_records(
            neo4j_records, node_classes, relationship_classes
        )

        return NeontologyResult(
            records_raw=neo4j_records,
            records=neontology_records,
            nodes=nodes,
            relationships=rels,
            paths=paths,
        )

    def evaluate_query_single(self, cypher: str, params: dict = {}) -> Optional[Any]:
        """Evaluate a Cypher query which returns a single result.

        Args:
            cypher (str): query to evaluate.
            params (dict, optional): parameters to pass through. Defaults to {}.

        Returns:
            Optional[Any]: Query result, or None if no result is found.
        """
        result = self.driver.execute_query(cypher, parameters_=params, result_transformer_=Neo4jResult.single)

        if result:
            return result.value()

        else:
            return None


class MemgraphConfig(GraphEngineConfig):
    """Configuration for a Neo4j graph engine."""

    engine: ClassVar[type[MemgraphEngine]] = MemgraphEngine
    uri: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None

    env_fields: ClassVar[dict[str, str]] = {
        "uri": "MEMGRAPH_URI",
        "username": "MEMGRAPH_USERNAME",
        "password": "MEMGRAPH_PASSWORD",
    }

    # Properties that guarantee non-None values for type checking
    @property
    def connection_uri(self) -> str:
        """Get the URI, guaranteed to be non-None after validation."""
        if self.uri is None:
            raise ValueError("URI should be set by validator")

        return self.uri

    @property
    def connection_username(self) -> str:
        """Get the username, guaranteed to be non-None after validation."""
        if self.username is None:
            raise ValueError("Username should be set by validator")

        return self.username

    @property
    def connection_password(self) -> str:
        """Get the password, guaranteed to be non-None after validation."""
        if self.password is None:
            raise ValueError("Password should be set by validator")

        return self.password



from pydantic import StringConstraints, TypeAdapter
from typing_extensions import Annotated

GQLIdentifier = Annotated[str, StringConstraints(strict=True, pattern=r"^[a-zA-Z][a-zA-Z0-9_]+$")]

# simple pydantic TypeAdapter for use validating strings before
# inserting them into GQL statements
gql_identifier_adapter = TypeAdapter(GQLIdentifier)

int_adapter = TypeAdapter(int)



from __future__ import annotations

from collections import defaultdict

from .basenode import BaseNode
from .baserelationship import BaseRelationship, RelationshipTypeData
from .graphconnection import GraphConnection


def get_node_types(
    base_type: type[BaseNode] = BaseNode,
) -> dict[str, type[BaseNode]]:
    """Get a dictionary of node types keyed by primary label.

    This is based on the class hierarchy of BaseNode and its subclasses.

    Optionally pass in a node class to only retrieve classes and subclasses of that type.

    Args:
        base_type (type[BaseNode], optional): Node type to return subclass information for. Defaults to BaseNode.

    Returns:
        dict[str, type[BaseNode]]: Dictionary of node types keyed by primary label.
    """
    node_types = {}

    # if we're starting with a node type that has a primary label, include this in results
    if getattr(base_type, "__primarylabel__", None):
        node_types[base_type.__primarylabel__] = base_type

    for subclass in base_type.__subclasses__():
        # we can define 'abstract' nodes which don't have a label
        # these are to provide common properties to be used by subclassed nodes
        # but shouldn't be put in the graph
        if getattr(subclass, "__primarylabel__", None):
            node_types[subclass.__primarylabel__] = subclass

        node_types.update(get_node_types(subclass))

    return node_types


def generate_relationship_type_data(
    rel_class: type[BaseRelationship],
) -> RelationshipTypeData:
    """Generate relationship type data for a given relationship class."""
    defined_source_class = rel_class.model_fields["source"].annotation
    defined_target_class = rel_class.model_fields["target"].annotation

    if not defined_source_class or not defined_target_class:
        raise ValueError(f"Relationship {rel_class.__name__} must have source and target classes defined.")

    all_source_classes = list(get_node_types(defined_source_class).values())
    all_target_classes = list(get_node_types(defined_target_class).values())

    return RelationshipTypeData(
        relationship_class=rel_class,
        source_class=defined_source_class,
        target_class=defined_target_class,
        all_source_classes=all_source_classes,
        all_target_classes=all_target_classes,
    )


def get_rels_by_type(
    base_type: type[BaseRelationship] = BaseRelationship,
) -> dict[str, RelationshipTypeData]:
    """Get a dictionary of relationship type data keyed by relationship type.

    Optionally pass in a relationship class to only retrieve classes and subclasses of that type.
    """
    rel_types: dict = defaultdict(dict)

    if getattr(base_type, "__relationshiptype__", None):
        rel_types[base_type.__relationshiptype__] = generate_relationship_type_data(base_type)

    for rel_subclass in base_type.__subclasses__():
        # we can define 'abstract' relationships which don't have a label
        # these are to provide common properties to be used by subclassed relationships
        # but shouldn't be put in the graph

        if getattr(rel_subclass, "__relationshiptype__", None):
            rel_types[rel_subclass.__relationshiptype__] = generate_relationship_type_data(rel_subclass)

        rel_types.update(get_rels_by_type(rel_subclass))

    return rel_types


def all_subclasses(cls: type) -> set:
    """Recursively get all subclasses of a given class."""
    subclasses = set(cls.__subclasses__())
    for subclass in cls.__subclasses__():
        subclasses.update(all_subclasses(subclass))
    return subclasses


def get_rels_by_node(base_type: type[BaseRelationship] = BaseRelationship, by_source: bool = True) -> dict[str, set[str]]:
    """Get a dictionary of relationship types by source or target node type."""
    if by_source is True:
        node_dir = "source_class"

    else:
        node_dir = "target_class"

    all_rels = get_rels_by_type(base_type)

    by_node: dict[str, set[str]] = defaultdict(set)

    for rel_type, entry in all_rels.items():
        try:
            node_label = entry.model_dump()[node_dir].__primarylabel__
        except AttributeError:
            node_label = None

        if node_label is not None:
            by_node[node_label].add(rel_type)

        for node_subclass in all_subclasses(entry.model_dump()[node_dir]):
            subclass_label = node_subclass.__primarylabel__
            if subclass_label is not None:
                by_node[subclass_label].add(rel_type)

    return by_node


def get_rels_by_source(
    base_type: type[BaseRelationship] = BaseRelationship,
) -> dict[str, set[str]]:
    """Get a dictionary of relationship types by source node type."""
    return get_rels_by_node(base_type, by_source=True)


def get_rels_by_target(
    base_type: type[BaseRelationship] = BaseRelationship,
) -> dict[str, set[str]]:
    """Get a dictionary of relationship types by target node type."""
    return get_rels_by_node(base_type, by_source=False)


def apply_neo4j_constraints(node_types: list[type[BaseNode]]) -> None:
    """Apply constraints based on primary properties for arbitrary set of node types."""
    graph = GraphConnection()

    for node_type in node_types:
        label = node_type.__primarylabel__
        if not label:
            raise ValueError("Node must have an explicit primary label to apply a constraint.")
        graph.engine.apply_constraint(label, node_type.__primaryproperty__)


def auto_constrain_neo4j() -> None:
    """Automatically apply constraints.

    Get information about all the defined nodes in the current environment.

    Apply constraints based on the primary label and primary property for each node.
    """
    node_types = list(get_node_types().values())

    apply_neo4j_constraints(node_types)



import json
from collections.abc import Generator, Sequence
from logging import getLogger
from pathlib import Path
from typing import Optional

import yaml

from .import_records import import_records

logger = getLogger(__name__)


def _identify_filepaths(input_path: str, path_pattern: str) -> list[Path]:
    path = Path(input_path)

    if path.is_file():
        return [path]

    return list(path.glob(path_pattern))


def _batch(iterable: Sequence, batch_size: int = 1) -> Generator:
    full_length = len(iterable)
    for ndx in range(0, full_length, batch_size):
        yield iterable[ndx : min(ndx + batch_size, full_length)]  # noqa: E203


def import_json(
    path: str,
    path_pattern: str = "**/*.json",
    batch_size: Optional[int] = None,
    check_unmatched: bool = True,
    error_on_unmatched: bool = False,
    validate_only: bool = False,
) -> None:
    """Import JSON files into the graph."""
    file_paths = _identify_filepaths(path, path_pattern)

    if len(file_paths) == 0:
        logger.warning("Didn't find any files to import")
        return

    if not batch_size:
        batch_size = len(file_paths)

    for file_batch in _batch(file_paths, batch_size):
        input_records = []

        for file_path in file_batch:
            logger.info("Processing %s", file_path)
            with open(file_path, "r") as json_file:
                raw_record_data = json.load(json_file)

            input_records.append(raw_record_data)

        import_records(
            input_records,
            check_unmatched=check_unmatched,
            error_on_unmatched=error_on_unmatched,
            validate_only=validate_only,
        )


def import_yaml(
    path: str,
    path_pattern: str = "**/*.yaml",
    batch_size: Optional[int] = None,
    check_unmatched: bool = True,
    error_on_unmatched: bool = False,
    validate_only: bool = False,
) -> None:
    """Import YAML files into the graph.

    The YAML files should contain a list of records, each record being a dictionary
    with LABEL for the primary label keys for other node properties.
    """
    file_paths = _identify_filepaths(path, path_pattern)

    if len(file_paths) == 0:
        logger.warning("Didn't find any files to import")
        return

    if batch_size is None:
        batch_size = len(file_paths)

    for file_batch in _batch(file_paths, batch_size):
        input_records = []

        for file_path in file_batch:
            logger.info("Processing %s", file_path)
            with open(file_path, "r") as yaml_file:
                raw_record_data = yaml.safe_load_all(yaml_file)

                input_records.append([x for x in raw_record_data])

        import_records(
            input_records,
            check_unmatched=check_unmatched,
            error_on_unmatched=error_on_unmatched,
            validate_only=validate_only,
        )


def import_md(
    path: str,
    path_pattern: str = "**/*.md",
    batch_size: Optional[int] = None,
    check_unmatched: bool = True,
    error_on_unmatched: bool = False,
    validate_only: bool = False,
) -> None:
    """Import markdown files with frontmatter into the graph.

    The frontmatter should be in YAML format and contain LABEL and a BODY_PROPERTY key
    """
    file_paths = _identify_filepaths(path, path_pattern)

    if len(file_paths) == 0:
        logger.warning("Didn't find any files to import")
        return

    if not batch_size:
        batch_size = len(file_paths)

    for file_batch in _batch(file_paths, batch_size):
        input_records = []

        for file_path in file_batch:
            logger.info("Processing %s", file_path)

            with open(file_path, "r") as md_file:
                raw_entry = md_file.read()

            entry = raw_entry.strip().split("---", maxsplit=2)

            frontmatter = entry[1]
            markdown_text = entry[2].strip()

            record = next(yaml.safe_load_all(frontmatter))
            body_property = record.pop("BODY_PROPERTY")

            record[body_property] = markdown_text

            input_records.append(record)

        import_records(
            input_records,
            check_unmatched=check_unmatched,
            error_on_unmatched=error_on_unmatched,
            validate_only=validate_only,
        )



from typing import ClassVar, Optional

from neontology import GraphConnection
from neontology.basenode import BaseNode
from neontology.baserelationship import BaseRelationship
from neontology.utils import (
    apply_neo4j_constraints,
    auto_constrain_neo4j,
    get_node_types,
    get_rels_by_source,
    get_rels_by_type,
)


def test_get_rels_by_source():
    class AbstractNodeType(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        pp: str

    class MyNodeType(AbstractNodeType):
        __primarylabel__: ClassVar[Optional[str]] = "MyNodeType"

    class MyRelType(BaseRelationship):
        source: AbstractNodeType
        target: MyNodeType
        __relationshiptype__: ClassVar[Optional[str]] = "MY_REL_TYPE"

    rels_by_source = get_rels_by_source()

    assert rels_by_source["MyNodeType"] == {"MY_REL_TYPE"}


def test_get_rels_by_type_subclasses():
    class MyNodeType1(BaseNode):
        __primarylabel__: ClassVar[Optional[str]] = "MyNodeType1"
        __primaryproperty__: ClassVar[str] = "pp"
        pp: str

    class MyAbstractRelType(BaseRelationship):
        source: MyNodeType1
        target: MyNodeType1

    class MyRelType1(MyAbstractRelType):
        __relationshiptype__: ClassVar[Optional[str]] = "MY_REL_TYPE1"

    class MyRelType2(MyAbstractRelType):
        __relationshiptype__: ClassVar[Optional[str]] = "MY_REL_TYPE2"

    class MyRelType3(BaseRelationship):
        source: MyNodeType1
        target: MyNodeType1
        __relationshiptype__: ClassVar[Optional[str]] = "MY_REL_TYPE_3"

    rels_by_type = get_rels_by_type(MyAbstractRelType)

    assert set(rels_by_type.keys()) == {"MY_REL_TYPE1", "MY_REL_TYPE2"}

    assert rels_by_type["MY_REL_TYPE1"].all_source_classes[0].__primarylabel__ == "MyNodeType1"


def test_no_primary_label_get_node_types():
    class SpecialPracticeLabelNode(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        pp: str

    node_types = get_node_types()

    assert isinstance(node_types, dict)

    assert "SpecialPracticeLabelNode" not in node_types.keys()


class SpecialPracticeNodeAC(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[str] = "SpecialPracticeNodeAC"
    pp: str


def test_auto_constrain_neo4j(use_graph):
    gc = GraphConnection()

    # make sure we start with no constraints

    try:
        constraints = gc.engine.get_constraints()

        for constraint_name in constraints:
            gc.engine.drop_constraint(constraint_name)

        result = gc.engine.get_constraints()

        assert len(result) == 0

        auto_constrain_neo4j()

        result2 = gc.engine.get_constraints()

        assert len(result2) >= 2

    # not all graph engines do constraints
    except NotImplementedError:
        pass


def test_apply_neo4j_constraints(use_graph):
    class SpecialPracticeNode(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[str] = "SpecialPracticeNode"
        pp: str

    gc = GraphConnection()

    try:
        # make sure we start with no constraints

        constraints = gc.engine.get_constraints()

        for constraint_name in constraints:
            gc.engine.drop_constraint(constraint_name)

        result = gc.engine.get_constraints()

        assert len(result) == 0

        apply_neo4j_constraints([SpecialPracticeNode])

        result2 = gc.engine.get_constraints()

        assert len(result2) == 1

    # not all graph engines do constraints
    except NotImplementedError:
        pass



from datetime import datetime
from typing import Optional
from uuid import UUID

import pytest
from pydantic import Field

from neontology.commonmodel import CommonModel


class PracticeModel(CommonModel):
    def _get_merge_parameters(self):
        return {}

    optional_string: Optional[str] = None


def test_common_model_creation():
    common = PracticeModel(optional_string="testing123")

    assert isinstance(common.optional_string, str)


def test_set_on_match():
    """Check that we successfully identify field to set on match"""

    class TestModel(PracticeModel):
        only_set_on_match: str = Field(json_schema_extra={"set_on_match": True})
        normal_field: str

    test_model = TestModel(only_set_on_match="Foo", normal_field="Bar")

    assert test_model._set_on_create == []
    assert test_model._set_on_match == ["only_set_on_match"]

    assert test_model._always_set == [
        "optional_string",
        "normal_field",
    ]


def test_set_on_create():
    """Check that we successfully identify field to set on match"""

    class TestModel(PracticeModel):
        only_set_on_create: str = Field(json_schema_extra={"set_on_create": True})
        normal_field: str

    test_model = TestModel(only_set_on_create="Foo", normal_field="Bar")

    assert test_model._set_on_create == ["only_set_on_create"]
    assert test_model._set_on_match == []
    assert test_model._always_set == [
        "optional_string",
        "normal_field",
    ]


@pytest.mark.parametrize(
    "field_type,python_value,neo4j_values",
    [
        (str, "hello world", ["hello world"]),
        (tuple, ("hello", "world"), [["hello", "world"]]),
        (set, {"foo", "bar"}, [["bar", "foo"], ["foo", "bar"]]),
        (
            UUID,
            UUID("32d4a4cb-29c3-4aa8-9b55-7790431819e3"),
            ["32d4a4cb-29c3-4aa8-9b55-7790431819e3"],
        ),
        (
            datetime,
            datetime(year=1984, month=1, day=2),
            [datetime(year=1984, month=1, day=2)],
        ),
        (
            list,
            ["foo", "bar"],
            [["foo", "bar"]],
        ),
    ],
)
def test_engine_dict(field_type, python_value, neo4j_values, use_graph):
    class TestModel(PracticeModel):
        test_prop: field_type

    testmodel = TestModel(test_prop=python_value)

    test_prop_result = testmodel._engine_dict()["test_prop"]

    assert test_prop_result in neo4j_values


@pytest.mark.parametrize(
    "field_type,python_value", [(dict, {"foo": "bar"}), (list, [123, "foo"])]
)
def test_engine_dict_bad_types(field_type, python_value, use_graph):
    class TestModel(PracticeModel):
        test_prop: field_type

    testmodel = TestModel(test_prop=python_value)

    with pytest.raises(TypeError):
        testmodel._engine_dict()



import functools
import json
import warnings
from typing import Any, Callable, ClassVar, Optional, TypeVar, Union

import pandas as pd
from pydantic import ValidationError, model_validator
from typing_extensions import ParamSpec, Self

from .commonmodel import CommonModel
from .gql import gql_identifier_adapter, int_adapter
from .graphconnection import GraphConnection
from .result import NeontologyResult
from .schema_utils import NodeSchema, SchemaProperty, extract_type_mapping

P = ParamSpec("P")
R = TypeVar("R")


def _find_this_node(query, params, node):
    this_node = f"(ThisNode:{node.__primarylabel__} {{{node.__primaryproperty__}: $_neontology_pp}})"
    params["_neontology_pp"] = node.get_pp()
    new_query = query.replace("(#ThisNode)", this_node)

    return new_query, params


def _prepare_related_query(node: "BaseNode", wrapped_function: Callable, *args: Any, **kwargs: Any) -> tuple[str, dict]:
    try:
        query, params = wrapped_function(node, *args, **kwargs)
    except ValueError:
        query = wrapped_function(node, *args, **kwargs)

        # if the function doesn't pass params, they may be taken from user provided parameters
        params = {**kwargs}

    # make it easy to match on this specific node
    if "(#ThisNode)" in query:
        new_query, params = _find_this_node(query, params, node)

    else:
        new_query = query

    return new_query, params


def related_property(f: Callable[P, R]) -> Callable:
    """Decorator to wrap functions on BaseNode subclasses and return a single result."""

    @functools.wraps(f)
    def wrapper(self: "BaseNode", *args: P.args, **kwargs: P.kwargs) -> Optional[Any]:
        new_query, params = _prepare_related_query(self, f, *args, **kwargs)

        gc = GraphConnection()
        result = gc.evaluate_query_single(new_query, params)

        return result

    wrapper.neontology_related_prop = True  # type: ignore

    return wrapper


def related_nodes(f: Callable[P, R]) -> Callable:
    """Decorator to wrap functions on BaseNode subclasses and return a list of nodes."""

    @functools.wraps(f)
    def wrapper(self: "BaseNode", *args: P.args, **kwargs: P.kwargs) -> list["BaseNode"]:
        new_query, params = _prepare_related_query(self, f, *args, **kwargs)

        gc = GraphConnection()
        result = gc.evaluate_query(new_query, params)

        return result.nodes

    wrapper.neontology_related_nodes = True  # type: ignore

    return wrapper


class BaseNode(CommonModel):  # pyre-ignore[13]
    __primaryproperty__: ClassVar[str]
    __primarylabel__: ClassVar[Optional[str]]
    __secondarylabels__: ClassVar[list[str]] = []

    def __init__(self, **data: dict):
        super().__init__(**data)

        # we can define 'abstract' nodes which don't have a label
        # these are to provide common properties to be used by subclassed nodes
        # but shouldn't be put in the graph or even instantiated
        if self.__primarylabel__ is None:
            raise NotImplementedError("Nodes to be used in the graph must define a primary label.")

    def __str__(self) -> str:
        """String representation of the node, showing the primary property value by default."""
        return str(self.get_pp())

    def _get_merge_parameters(self) -> dict[str, Any]:
        """Get the parameters defined for merging on when merging into the graph.

        Returns:
        dict[str, Any]: a dictionary of key/value pairs.
        """
        params = self._get_merge_parameters_common()
        # get all the properties
        all_props = params.pop("all_props")

        params.update(
            {
                "pp": all_props[self.__primaryproperty__],
            }
        )

        return params

    @classmethod
    def get_related_node_methods(cls) -> dict:
        """Gather all methods tagged with `related_nodes` decorator.

        Returns:
            dict: A dictionary of method names and their corresponding methods.
        """
        related_node_attributes = {
            name: getattr(cls, name)
            # get all attributes, including methods, properties, and builtins
            for name in dir(cls)
            # only want methods
            if callable(getattr(cls, name))
            # filter to tagged methods
            and hasattr(getattr(cls, name), "neontology_related_nodes")
        }
        return related_node_attributes

    @classmethod
    def get_related_property_methods(cls) -> dict:
        """Gather all methods tagged with `related_property` decorator.

        Returns:
            dict: A dictionary of method names and their corresponding methods.
        """
        related_prop_attributes = {
            name: getattr(cls, name)
            # get all attributes, including methods, properties, and builtins
            for name in dir(cls)
            # only want methods
            if callable(getattr(cls, name))
            # filter to tagged methods
            and hasattr(getattr(cls, name), "neontology_related_prop")
        }
        return related_prop_attributes

    @model_validator(mode="after")
    def validate_identifiers(self) -> Self:
        """Validate data provided for primary label and primary property."""
        try:
            gql_identifier_adapter.validate_strings(self.__primarylabel__)

        except AttributeError:
            pass
        except ValidationError:
            warnings.warn(
                (
                    "Primary Label should contain only alphanumeric characters and underscores."
                    " It should begin with an alphabetic character."
                )
            )

        try:
            gql_identifier_adapter.validate_strings(self.__primaryproperty__)
        except AttributeError:
            pass
        except ValidationError:
            warnings.warn(
                (
                    "Primary Property should contain only alphanumeric characters and underscores."
                    " It should begin with an alphabetic character."
                )
            )

        return self

    def get_pp(self) -> Union[str, int]:
        """Get the primary property value for this node.

        Returns:
            Union[str, int]: The value of the primary property.
        """
        return self._get_merge_parameters()["pp"]

    def get_primary_property_value(self) -> Union[str, int]:
        """Get the primary property value for this node.

        Deprecated: Use `get_pp()` instead.

        Returns:
            Union[str, int]: The value of the primary property.
        """
        warnings.warn(("get_primary_property_value is deprecated, use get_pp instead."))
        return self.get_pp()

    def create(self) -> Self:
        """Create this node in the graph."""
        all_props = self._engine_dict()

        pp_value = all_props.pop(self.__primaryproperty__)

        node_details = [{"pp": pp_value, "props": all_props}]

        all_labels = [self.__primarylabel__] + self.__secondarylabels__

        pp_key = self.__primaryproperty__

        gc = GraphConnection()

        results = gc.create_nodes(all_labels, pp_key, node_details, self.__class__)

        return results[0]

    def merge(self) -> list[Self]:
        """Merge this node into the graph."""
        node_list = [self._get_merge_parameters()]

        all_labels = [self.__primarylabel__] + self.__secondarylabels__

        pp_key = self.__primaryproperty__

        gc = GraphConnection()

        results = gc.merge_nodes(all_labels, pp_key, node_list, self.__class__)

        return results

    @classmethod
    def create_nodes(cls, nodes: list[Self]) -> list[Self]:
        """Create the given nodes in the database.

        Args:
            nodes (list[B]): A list of nodes to create.

        Returns:
            list: A list of the primary property values

        Raises:
            TypeError: Raised if one of the nodes isn't of this type.
        """
        node_list = [{"props": x._engine_dict(), "pp": x._engine_dict()[cls.__primaryproperty__]} for x in nodes]

        all_labels = [cls.__primarylabel__] + cls.__secondarylabels__
        pp_key = cls.__primaryproperty__

        gc = GraphConnection()

        results = gc.create_nodes(all_labels, pp_key, node_list, cls)

        return results

    @classmethod
    def merge_nodes(cls, nodes: list[Self]) -> list[Self]:
        """Merge multiple nodes into the database.

        Args:
            nodes (list[B]): A list of nodes to merge.

        Returns:
            list: A list of the primary property values

        Raises:
            TypeError: Raised if any of the nodes provided don't match this class.
        """
        node_list = [x._get_merge_parameters() for x in nodes]

        all_labels = [cls.__primarylabel__] + cls.__secondarylabels__

        pp_key = cls.__primaryproperty__

        gc = GraphConnection()

        results = gc.merge_nodes(all_labels, pp_key, node_list, cls)

        return results

    @classmethod
    def merge_records(cls, records: list[dict]) -> list[Self]:
        """Take a list of dictionaries and use them to merge in nodes in the graph.

        Each dictionary will be used to merge a node where dictionary key/value pairs
            represent properties to be applied.

        Returns:
            list: A list of the primary property values

        Args:
            records (list[dict[str, Any]]): a list of dictionaries of node properties
        """
        nodes = [cls(**x) for x in records]

        return cls.merge_nodes(nodes)

    @classmethod
    def merge_df(cls, df: pd.DataFrame, deduplicate: bool = True) -> pd.Series:
        """Merge in new nodes based on data in a dataframe.

        The dataframe columns must correspond to the Node properties.

        Returns:
            pd.Series: A list of the primary property values

        Args:
            df (pd.DataFrame): A pandas dataframe of node properties
            deduplicate (bool): If True, deduplicate the dataframe before merging.
                Defaults to True.

        """
        if df.empty is True:
            return pd.Series(dtype=object)

        input_df = df.mask(pd.isna(df), None).copy()

        # create a unique identifier field based on all rows
        # we'll use this later to match up deduplicated rows to the original ordering
        input_df["unique_identifier"] = input_df.astype(str).values.sum(axis=1)

        if deduplicate is True:
            # we don't wan't to waste time attempting to merge identical records
            unique_df = input_df.drop_duplicates(subset="unique_identifier", ignore_index=True).copy()
        else:
            unique_df = input_df

        model_data = unique_df.drop("unique_identifier", axis=1).copy()

        records = model_data.to_dict(orient="records")

        unique_df["generated_nodes"] = pd.Series(cls.merge_records(records))

        # now we need to get the mapping from unique id to primary property
        # so that we can return the data in the same shape it was received
        input_df.insert(0, "ontolocy_merging_order", range(0, len(input_df)))
        output_df = input_df.merge(unique_df, how="outer", on="unique_identifier", suffixes=(None, "_y"))

        ordered_nodes = output_df.sort_values("ontolocy_merging_order", ignore_index=True).generated_nodes.copy()

        return ordered_nodes

    @classmethod
    def match(cls, pp: str) -> Optional[Self]:
        """MATCH a single node of this type with the given primary property.

        Args:
            pp (str): The value of the primary property (pp) to match on.

        Returns:
            Optional[B]: If the node exists, return it as an instance.
        """
        cypher = f"""
        MATCH (n:{cls.__primarylabel__})
        WHERE n.{cls.__primaryproperty__} = $pp
        RETURN n
        """

        params = {"pp": pp}

        gc = GraphConnection()

        result = gc.evaluate_query(cypher, params, node_classes={cls.__primarylabel__: cls})

        if result.nodes:
            return result.nodes[0]

        else:
            return None

    @classmethod
    def delete(cls, pp: str) -> None:
        """Delete a node from the graph.

        Match on label and the pp value provided.
        If the node exists, delete it and any relationships it has.

        Args:
            pp (str): Primary property value to match on.
        """
        label = cls.__primarylabel__

        if label is None:
            raise ValueError("Cannot delete nodes without a primary label.")

        pp_key = cls.__primaryproperty__
        pp_value = pp

        gc = GraphConnection()

        gc.delete_nodes(label, pp_key, [pp_value])

    @classmethod
    def match_nodes(cls, limit: Optional[int] = None, skip: Optional[int] = None) -> list[Self]:
        """Get nodes of this type from the database.

        Run a MATCH cypher query to retrieve any Nodes with the label of this class.

        Args:
            limit (int, optional): Maximum number of results to return. Defaults to None.
            skip (int, optional): Skip through this many results (for pagination). Defaults to None.

        Returns:
            Optional[list[B]]: A list of node instances.
        """
        gc = GraphConnection()
        result = gc.match_nodes(cls, limit, skip)

        return result

    def get_related(
        self,
        relationship_types: list = [],
        relationship_properties: Optional[dict] = None,
        target_label: Optional[str] = None,
        outgoing: bool = True,
        incoming: bool = False,
        depth: Optional[tuple] = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
        distinct: bool = False,
    ) -> NeontologyResult:
        """Get related nodes based on relationships of this node.

        Args:
            relationship_types (list, optional): List of relationship types to match on.
                Defaults to [].
            relationship_properties (dict, optional): Dictionary of relationship properties to match on.
                Defaults to None.
            target_label (str, optional): Label of the target node to match.
                Defaults to None, which matches any label.
            outgoing (bool, optional): If True, match outgoing relationships.
                Defaults to True.
            incoming (bool, optional): If True, match incoming relationships.
                Defaults to False.
            depth (tuple, optional): A tuple of (min_depth, max_depth) to limit the depth of relationships.
                Defaults to None, which matches any depth.
            limit (int, optional): Maximum number of results to return. Defaults to None.
            skip (int, optional): Skip through this many results (for pagination). Defaults to None.
            distinct (bool, optional): If True, return distinct results. Defaults to False.

        Returns:
            NeontologyResult: A result object containing the nodes and relationships found.

        Raises:
            ValueError: If neither outgoing nor incoming is specified.
        """
        if target_label:
            target = f"o:{gql_identifier_adapter.validate_strings(target_label)}"
        else:
            target = "o"

        if relationship_types:
            rel_type_match = "r:" + "|".join([gql_identifier_adapter.validate_strings(x) for x in relationship_types])

        else:
            rel_type_match = "r"

        if relationship_properties:
            rel_prop_match = (
                "{" + ", ".join([f"{gql_identifier_adapter.validate_strings(x)}: ${x}" for x in relationship_properties]) + "}"
            )

            pass_on_params = dict(relationship_properties)

        else:
            rel_prop_match = ""
            pass_on_params = {}

        if outgoing and incoming:
            out_dir = "-"
            in_dir = "-"

        elif outgoing:
            out_dir = "->"
            in_dir = "-"

        elif not outgoing and not incoming:
            raise ValueError("Must specify at least one of incoming or outgoing.")

        else:
            out_dir = "-"
            in_dir = "<-"

        if depth:
            min_depth, max_depth = depth

            rel_depth = f"*{min_depth}..{max_depth}"
        else:
            rel_depth = ""

        if distinct:
            return_distinct = "DISTINCT"

        else:
            return_distinct = ""

        query = f"""
        MATCH (#ThisNode){in_dir}[{rel_type_match}{rel_depth} {rel_prop_match}]{out_dir}({target})
        RETURN {return_distinct} o, r, ThisNode
        """

        if skip:
            query += f" SKIP {int_adapter.validate_python(skip)} "

        if limit:
            query += f" LIMIT {int_adapter.validate_python(limit)} "

        new_query, params = _find_this_node(query, pass_on_params, self)

        gc = GraphConnection()
        result = gc.evaluate_query(new_query, params)

        return result

    @classmethod
    @related_property
    def get_count(cls):
        """Get the count of nodes of this type in the graph."""
        return f"MATCH (n:{cls.__primarylabel__}) RETURN COUNT(DISTINCT n)"

    def _prep_dump_dict(self, dumped_model: dict) -> dict:
        dumped_model["LABEL"] = self.__primarylabel__

        return dumped_model

    def neontology_dump(self, exclude: Optional[set] = None, exclude_none: bool = True, **kwargs) -> dict:
        """Dump the model as a dictionary which can be reimported.

        Args:
            exclude (set, optional): A set of fields to exclude from the dump.
                Defaults to None.
            exclude_none (bool, optional): If True, exclude fields with None values.
                Defaults to True.
            **kwargs: Additional keyword arguments to pass to the model_dump method.

        Returns:
            dict: A dictionary representation of the model.
        """
        dumped_model = self.model_dump(exclude_none=exclude_none, exclude=exclude, **kwargs)

        return self._prep_dump_dict(dumped_model)

    def neontology_dump_json(self, exclude: Optional[set] = None, exclude_none: bool = True, **kwargs) -> str:
        """Dump the model as a JSON string which can be reimported.

        Args:
            exclude (set, optional): A set of fields to exclude from the dump.
                Defaults to None.
            exclude_none (bool, optional): If True, exclude fields with None values.
                Defaults to True.
            **kwargs: Additional keyword arguments to pass to the model_dump_json method.

        Returns:
            str: A JSON string representation of the model.
        """
        # pydantic converts values to be json serializable, make use of this first
        original_json = self.model_dump_json(exclude_none=exclude_none, exclude=exclude, **kwargs)
        model_dict = json.loads(original_json)
        model_dict["LABEL"] = self.__primarylabel__
        return json.dumps(self._prep_dump_dict(model_dict))

    @classmethod
    def neontology_schema(cls, include_outgoing_rels: bool = True) -> NodeSchema:
        """Generate a schema for this node class.

        Args:
            include_outgoing_rels (bool, optional): If True, include outgoing relationships in the schema.
                Defaults to True.

        Returns:
            NodeSchema: A schema object representing the node class.

        Raises:
            ValueError: If the node class does not have a primary label defined.
        """
        if not cls.__primarylabel__:
            raise ValueError("Node does not have a primary label defined for generating schema.")

        schema_dict: dict = {}
        schema_dict["label"] = cls.__primarylabel__
        schema_dict["title"] = cls.__name__
        schema_dict["secondary_labels"] = cls.__secondarylabels__

        model_properties: list = []

        for field_name, field_props in cls.model_fields.items():
            field_type = extract_type_mapping(field_props.annotation, show_optional=True)

            node_property = SchemaProperty(
                type_annotation=field_type,
                name=field_name,
                required=field_props.is_required(),
            )

            if field_props.is_required() is True:
                model_properties.insert(0, node_property)

            # put optional fields at the end
            else:
                model_properties.append(node_property)

        schema_dict["properties"] = model_properties
        schema_dict["outgoing_relationships"] = []

        if include_outgoing_rels is False:
            return NodeSchema(**schema_dict)

        else:
            from .utils import get_rels_by_source, get_rels_by_type

            outgoing_rels = get_rels_by_source().get(cls.__primarylabel__, set())
            all_rel_types = get_rels_by_type()

            for rel in outgoing_rels:
                rel_class = all_rel_types[rel].relationship_class

                rel_schema = rel_class.neontology_schema(source_labels=[schema_dict["label"]])

                schema_dict["outgoing_relationships"].append(rel_schema)

        return NodeSchema(**schema_dict)



# flake8: noqa

from .basenode import BaseNode, related_nodes, related_property
from .baserelationship import BaseRelationship
from .gql import GQLIdentifier, gql_identifier_adapter
from .graphconnection import GraphConnection, init_neontology
from .graphengines.memgraphengine import MemgraphConfig
from .graphengines.neo4jengine import Neo4jConfig
from .utils import auto_constrain_neo4j

__all__ = [
    # BaseNode
    "BaseNode",
    "related_nodes",
    "related_property",
    # BaseRelationship
    "BaseRelationship",
    # GraphConnection
    "init_neontology",
    "GraphConnection",
    # utils
    "auto_constrain_neo4j",
    # GQL
    "GQLIdentifier",
    "gql_identifier_adapter",
    # Engines
    "Neo4jConfig",
    "MemgraphConfig",
]



from __future__ import annotations

import os
from datetime import date, datetime, time, timedelta
from typing import TYPE_CHECKING, Any, ClassVar, Optional, TypeVar

from dotenv import load_dotenv
from pydantic import BaseModel, model_validator

from ..gql import gql_identifier_adapter, int_adapter
from ..result import NeontologyResult

if TYPE_CHECKING:
    from ..basenode import BaseNode
    from ..baserelationship import BaseRelationship

BaseNodeT = TypeVar("BaseNodeT", bound="BaseNode")
BaseRelationshipT = TypeVar("BaseRelationshipT", bound="BaseRelationship")


class GraphEngineBase:
    _supported_types: ClassVar[Any] = (
        list,
        bool,
        int,
        bytearray,
        float,
        str,
        bytes,
        date,
        time,
        datetime,
        timedelta,
    )

    def __init__(self, config: Optional["GraphEngineConfig"]) -> None:
        """Initialise the graph engine.

        Args:
            config (Optional[dict]): GraphEngine configuration
        """
        pass

    @classmethod
    def _export_type_converter(cls, value: Any) -> Any:
        """Convert a value to a type supported by the graph engine.

        This method is used to ensure that values are converted to types that the graph engine can handle.

        Args:
            value (Any): The value to convert.

        Returns:
            Any: The converted value, or the original value if it is already of a supported type.

        Raises:
            TypeError: If the value is a dict, or if it is a list with mixed types.
        """
        if isinstance(value, dict):
            raise TypeError("Neontology doesn't support dict types for properties.")

        elif isinstance(value, (tuple, set)):
            new_value = list(value)
            return cls._export_type_converter(new_value)

        elif isinstance(value, list):
            if len(value) == 0:
                return []
            # items in a list must all be the same type
            item_type = type(value[0])
            for item in value:
                if isinstance(item, item_type) is False:
                    raise TypeError("For neo4j, all items in a list must be of the same type.")

            return [cls._export_type_converter(x) for x in value]

        elif value is None:
            return None

        elif isinstance(value, cls._supported_types) is False:
            return str(value)

        else:
            return value

    @classmethod
    def export_dict_converter(cls, original_dict: dict[str, Any]) -> dict[str, Any]:
        """Convert types in a dictionary to those supported by the graph engine.

        Args:
            original_dict (dict[str, Any]): The original dictionary to convert.

        Returns:
            dict[str, Any]: A new dictionary with values converted to types supported by the graph engine.
        """
        export_dict = original_dict.copy()

        for k, v in export_dict.items():
            export_dict[k] = cls._export_type_converter(v)

        return export_dict

    def verify_connection(self) -> bool:
        """Verify the connection to the graph engine.

        Returns:
            bool: True if the connection is valid, False otherwise.
        """
        raise NotImplementedError

    def close_connection(self) -> None:
        """Close the connection to the graph engine."""
        raise NotImplementedError

    def evaluate_query(
        self,
        cypher: str,
        params: dict[str, Any] = {},
        node_classes: dict = {},
        relationship_classes: dict = {},
    ) -> NeontologyResult:
        """Evaluate a Cypher query against the database.

        Args:
            cypher (str): The Cypher query to execute.
            params (dict[str, Any], optional): Parameters for the Cypher query. Defaults to {}.
            node_classes (dict, optional): Mapping of node labels to their classes. Defaults to {}.
            relationship_classes (dict, optional): Mapping of relationship types to their classes. Defaults to {}.

        Returns:
            NeontologyResult: The result of the query execution, containing nodes and relationships.
        """
        raise NotImplementedError

    def evaluate_query_single(self, cypher: str, params: dict[str, Any]) -> Any:
        """Evaluate a query which returns a single result.

        Args:
            cypher (str): The Cypher query to execute.
            params (dict[str, Any]): Parameters for the Cypher query.

        Returns:
            Any: The result of the query execution.
        """
        raise NotImplementedError

    def apply_constraint(self, label: str, property: str) -> None:
        """Apply a constraint to a label and property in the database.

        Args:
            label (str): The label to apply the constraint to.
            property (str): The property to apply the constraint on.
        """
        raise NotImplementedError

    def drop_constraint(self, constraint_name: str) -> None:
        """Drop a constraint from the database.

        Args:
            constraint_name (str): The name of the constraint to drop.
        """
        raise NotImplementedError

    def get_constraints(self) -> list:
        """Get a list of constraints in the database.

        Returns:
            list: A list of constraints in the database.
        """
        raise NotImplementedError

    def create_nodes(self, labels: list, pp_key: str, properties: list, node_class: type[BaseNodeT]) -> list[BaseNodeT]:
        """Create nodes with specified labels and properties.

        Args:
            labels (list): a list of labels to give created nodes
            pp_key (str): the primary property for the nodes
            properties (list): A list of dictionaries representing each node to be created.
                two keys with associated values pp (the value to assign the primary property)
                and props (dict with key value pairs for all other properties).
            node_class (type[BaseNodeT]): the type of nodes to create

        Returns:
            list: list of created Nodes
        """
        label_identifiers = [gql_identifier_adapter.validate_strings(x) for x in labels]

        cypher = f"""
        UNWIND $node_list AS node
        create (n:{":".join(label_identifiers)} {{{gql_identifier_adapter.validate_strings(pp_key)}: node.pp}})
        SET n += node.props
        RETURN n
        """

        params = {"node_list": properties}

        node_classes = {node_class.__primarylabel__: node_class}

        results = self.evaluate_query(cypher, params, node_classes)

        return results.nodes

    def merge_nodes(self, labels: list, pp_key: str, properties: list, node_class: type[BaseNodeT]) -> list[BaseNodeT]:
        """Merge nodes with specified labels and property.

        Args:
            labels (list): _description_
            pp_key (str): _description_
            properties (list): A list of dictionaries representing each node to be created.
                four keys with associated values: pp (the value to assign the primary property)
                set_on_match, set_on_create and always_set (dicts with key value pairs for all other properties).
            node_class (type["BaseNode"]): The class of the nodes to be merged.

        Returns:
            list: list of merged Nodes
        """
        label_identifiers = [gql_identifier_adapter.validate_strings(x) for x in labels]

        cypher = f"""
        UNWIND $node_list AS node
        MERGE (n:{":".join(label_identifiers)} {{{gql_identifier_adapter.validate_strings(pp_key)}: node.pp}})
        ON MATCH SET n += node.set_on_match
        ON CREATE SET n += node.set_on_create
        SET n += node.always_set
        RETURN n
        """

        params = {"node_list": properties}

        node_classes = {node_class.__primarylabel__: node_class}

        results = self.evaluate_query(cypher, params, node_classes)

        return results.nodes

    def delete_nodes(self, label: str, pp_key: str, pp_values: list[Any]) -> None:
        """Delete nodes with a specific label and primary property value.

        Args:
            label (str): The label of the nodes to delete.
            pp_key (str): The primary property key to match on.
            pp_values (list[Any]): A list of primary property values to match on for deletion.
        """
        cypher = f"""
        UNWIND $pp_values AS pp
        MATCH (n:{gql_identifier_adapter.validate_strings(label)})
        WHERE n.{gql_identifier_adapter.validate_strings(pp_key)} = pp
        DETACH DELETE n
        """

        params = {"pp_values": pp_values}

        self.evaluate_query_single(cypher, params)

    def merge_relationships(
        self,
        source_label: str,
        target_label: str,
        source_prop: str,
        target_prop: str,
        rel_type: str,
        merge_on_props: list[str],
        rel_props: list[dict],
    ) -> None:
        """Merge relationships between nodes in the database.

        Args:
            source_label (str): The label of the source node.
            target_label (str): The label of the target node.
            source_prop (str): The property of the source node to match on.
            target_prop (str): The property of the target node to match on.
            rel_type (str): The type of relationship to create or merge.
            merge_on_props (list[str]): A list of properties to merge on.
            rel_props (list[dict]): A list of dictionaries representing each relationship to be merged.
                Each dictionary should contain keys for `source_prop`, `target_prop`, and any additional properties
                to set on the relationship.
        """
        # build a string of properties to merge on "prop_name: $prop_name"
        merge_props = ", ".join([f"{gql_identifier_adapter.validate_strings(x)}: rel.{x}" for x in merge_on_props])

        cypher = f"""
        UNWIND $rel_list AS rel
        MATCH (source:{gql_identifier_adapter.validate_strings(source_label)})
        WHERE source.{gql_identifier_adapter.validate_strings(source_prop)} = rel.source_prop
        MATCH (target:{gql_identifier_adapter.validate_strings(target_label)})
        WHERE target.{gql_identifier_adapter.validate_strings(target_prop)} = rel.target_prop
        MERGE (source)-[r:{gql_identifier_adapter.validate_strings(rel_type)} {{ {merge_props} }}]->(target)
        ON MATCH SET r += rel.set_on_match
        ON CREATE SET r += rel.set_on_create
        SET r += rel.always_set
        """

        params = {"rel_list": rel_props}

        self.evaluate_query_single(cypher, params)

    def match_nodes(
        self,
        node_class: type[BaseNodeT],
        limit: Optional[int] = None,
        skip: Optional[int] = None,
    ) -> list[BaseNodeT]:
        """Get nodes of this type from the database.

        Run a MATCH cypher query to retrieve any Nodes with the label of this class.

        Args:
            node_class (type[BaseNodeT]): the type (primary label) of nodes to match on.
            limit (int, optional): Maximum number of results to return. Defaults to None.
            skip (int, optional): Skip through this many results (for pagination). Defaults to None.

        Returns:
            Optional[list[B]]: A list of node instances.
        """
        cypher = f"""
        MATCH(n:{node_class.__primarylabel__})
        RETURN n
        """

        params = {}

        if skip:
            cypher += " SKIP $skip "
            params["skip"] = int_adapter.validate_python(skip)

        if limit:
            cypher += " LIMIT $limit "
            params["limit"] = int_adapter.validate_python(limit)

        result = self.evaluate_query(cypher, params, node_classes={node_class.__primarylabel__: node_class})

        return result.nodes

    def match_relationships(
        self,
        relationship_class: type[BaseRelationshipT],
        limit: Optional[int] = None,
        skip: Optional[int] = None,
    ) -> list[BaseRelationshipT]:
        """Get relationships of this type from the database.

        Run a MATCH cypher query to retrieve any Relationships of this type.

        Args:
            relationship_class (type["BaseRelationshipT"]): the type of relationship to match on
            limit (int, optional): Maximum number of results to return. Defaults to None.
            skip (int, optional): Skip through this many results (for pagination). Defaults to None.

        Returns:
            list["BaseRelationshipT"]: A list of relationships.
        """
        from ..utils import get_node_types, get_rels_by_type

        cypher = f"""
        MATCH (n)-[r:{relationship_class.__relationshiptype__}]->(o)
        RETURN DISTINCT n, r, o
        """

        params = {}

        if skip:
            cypher += " SKIP $skip "
            params["skip"] = int_adapter.validate_python(skip)

        if limit:
            cypher += " LIMIT $limit "
            params["limit"] = int_adapter.validate_python(limit)

        rel_types = get_rels_by_type()
        node_classes = get_node_types()

        result = self.evaluate_query(
            cypher,
            params,
            node_classes=node_classes,
            relationship_classes=rel_types,
        )

        return result.relationships


class GraphEngineConfig(BaseModel):
    """Base class for Graph Engine configuration."""

    engine: ClassVar[type[GraphEngineBase]]

    env_fields: ClassVar[dict[str, str]]

    @model_validator(mode="before")
    @classmethod
    def populate_defaults(cls, data: Any) -> Any:
        """Populate configuration with environment variables.

        Where no values are provided, attempt to load them from environment variables.
        """
        load_dotenv()

        for field, env_var in cls.env_fields.items():
            if not data.get(field):
                value = os.getenv(env_var)
                if value is None:
                    raise ValueError(f"No value provided for {field} field and no {env_var} environment variable set.")
                data[field] = value
        return data



from typing import Any

from pydantic import BaseModel, ConfigDict, PrivateAttr

from .graphconnection import GraphConnection


class CommonModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        extra="forbid",
        arbitrary_types_allowed=True,
    )

    _set_on_match: list[str] = PrivateAttr()
    _set_on_create: list[str] = PrivateAttr()
    _always_set: list[str] = PrivateAttr()

    def __init__(self, **data: dict):
        super().__init__(**data)
        self._set_prop_usage()

    @classmethod
    def _set_prop_usage(cls) -> None:
        """Set the properties that are used by Neontology for specific purposes.

        This method initializes the class attributes `_set_on_match`, `_set_on_create`, and `_always_set`
        based on the model's JSON schema.
        It retrieves properties that are marked for specific usage types such as 'set_on_match' and 'set_on_create'.
        It also sets the `_always_set` attribute to include properties that are not in `_set_on_match` or `_set_on_create`
        and are not 'source' or 'target'.
        """
        cls._set_on_match = cls._get_prop_usage("set_on_match")
        cls._set_on_create = cls._get_prop_usage("set_on_create")
        cls._always_set = [
            v.alias if v.alias else x
            for x, v in cls.model_fields.items()
            if x not in cls._set_on_match + cls._set_on_create + ["source", "target"]
        ]

    @classmethod
    def _get_prop_usage(cls, usage_type: str) -> list[str]:
        """Get a list of properties that are used by Neontology for a specific purpose.

        These enable complex creation and merging use cases based on model metadata.

        Args:
            usage_type (str): The type of usage to filter properties by.
                              Can be 'set_on_match', 'set_on_create', or 'always_set'.

        Returns:
            list[str]: A list of property names that match the specified usage type.
        """
        all_props = cls.model_json_schema()["properties"]

        selected_props = []

        for prop, entry in all_props.items():
            if entry.get(usage_type) is True:
                selected_props.append(prop)

        return selected_props

    def _get_prop_values(
        self, props: list[str], exclude: set[str] = set()
    ) -> dict[str, Any]:
        """Get a dictionary of property values for the given properties.

        Args:
        props (list[str]): List of property names to include in the output.
        exclude (set[str], optional): Properties to exclude from the output. Defaults to set().

        Returns:
            dict[str, Any]: a dictionary of key/value pairs.
        """
        # prop_values = {
        #    k: v for k, v in self._engine_dict(exclude=exclude).items() if k in props
        # }

        return self._engine_dict(exclude=exclude, include=set(props))

    def _engine_dict(self, exclude: set[str] = set(), **kwargs: Any) -> dict[str, Any]:
        """Return a dict made up of only types compatible with the GraphEngine.

        Args:
            exclude (set[str], optional): Properties to exclude from the output. Defaults to set().
            **kwargs: Additional keyword arguments to pass to the model_dump method.

        Returns:
            dict: a dictionary export of this model instance
        """
        pydantic_export_dict = self.model_dump(
            exclude_none=False, exclude=exclude, by_alias=True, **kwargs
        )

        # return pydantic_export_dict

        try:
            gc = GraphConnection()
            export_dict = gc.engine.export_dict_converter(pydantic_export_dict)

        except RuntimeError:
            export_dict = pydantic_export_dict

        return export_dict

    def _get_merge_parameters_common(self, exclude: set[str] = set()) -> dict[str, Any]:
        """Input an all properties dictionary, and filter based on property types.

        Args:
            exclude (set[str], optional): Properties to exclude from the output. Defaults to set().

        Returns:
            Dict[str, Any]: Dictionary of always_set, set_on_match, and set_on_create dictionaries
        """
        # get all the properties
        all_props = self._engine_dict(exclude=exclude)

        always_set = {k: all_props[k] for k in self._always_set}
        set_on_match = {k: all_props[k] for k in self._set_on_match}
        set_on_create = {k: all_props[k] for k in self._set_on_create}
        params = {
            "all_props": all_props,
            "always_set": always_set,
            "set_on_match": set_on_match,
            "set_on_create": set_on_create,
        }
        return params



import itertools
import warnings
from typing import TYPE_CHECKING, Any, ClassVar, Optional, TypeVar

from neo4j import GraphDatabase
from neo4j import Record as Neo4jRecord
from neo4j import Result as Neo4jResult
from neo4j.graph import Node as Neo4jNode
from neo4j.graph import Path as Neo4jPath
from neo4j.graph import Relationship as Neo4jRelationship
from neo4j.time import Date as Neo4jDate
from neo4j.time import DateTime as Neo4jDateTime
from neo4j.time import Time as Neo4jTime
from typing_extensions import LiteralString, cast

from ..gql import gql_identifier_adapter
from ..result import NeontologyResult
from .graphengine import GraphEngineBase, GraphEngineConfig

if TYPE_CHECKING:
    from ..basenode import BaseNode
    from ..baserelationship import BaseRelationship, RelationshipTypeData

BaseNodeT = TypeVar("BaseNodeT", bound="BaseNode")


def convert_neo4j_types(input_dict: dict) -> dict:
    """Convert Neo4j types in a dictionary to their native Python equivalents.

    Specifically, this function converts Neo4j DateTime, Date, and Time
    objects to their native Python types.

    Args:
        input_dict (dict): Dictionary containing Neo4j types.

    Returns:
        dict: Dictionary with Neo4j types converted to native Python types.
    """
    output_dict = dict(input_dict)

    for key in output_dict:
        try:
            new_date = output_dict[key].to_native()

            if isinstance(output_dict[key], (Neo4jDateTime, Neo4jDate, Neo4jTime)):
                output_dict[key] = new_date

        except AttributeError:
            pass

    return output_dict


def neo4j_node_to_neontology_node(neo4j_node: Neo4jNode, node_classes: dict[str, type[BaseNodeT]]) -> Optional[BaseNodeT]:
    """Convert a native Neo4j node to a Neontology node.

    Args:
        neo4j_node (Neo4jNode): The Neo4j node to convert.
        node_classes (dict): Mapping of labels to node classes used for populating with results.

    Returns:
        Optional[BaseNode]: The converted Neontology node, or None if conversion fails.
    """
    node_labels = list(neo4j_node.labels)

    primary_labels = set(node_labels).intersection(set(node_classes.keys()))

    secondary_labels = set(node_labels).difference(set(node_classes.keys()))

    if len(primary_labels) == 1:
        primary_label = primary_labels.pop()

        node_dict = convert_neo4j_types(dict(neo4j_node))

        node = node_classes[primary_label](**node_dict)

        # warn if the secondary labels aren't what's expected

        if set(node.__secondarylabels__) != secondary_labels:
            warnings.warn(f"Unexpected secondary labels returned: {secondary_labels}")

        return node

    # gracefully handle cases where we don't have a class defined
    # for the identified label or where we get more than one valid primary label
    else:
        warnings.warn(f"Unexpected primary labels returned: {primary_labels}")

        return None


def neo4j_relationship_to_neontology_rel(
    neo4j_rel: Neo4jRelationship, node_classes: dict, rel_classes: dict
) -> Optional["BaseRelationship"]:
    """Convert a native Neo4j relationship to a Neontology relationship.

    Args:
        neo4j_rel (Neo4jRelationship): The Neo4j relationship to convert.
        node_classes (dict): Mapping of labels to node classes used for populating with results.
        rel_classes (dict[str, RelationshipTypeData]): Mapping of relationship types to classes for populating with results.

    Returns:
        Optional[BaseRelationship]: The converted Neontology relationship, or None if conversion fails.
    """
    rel_type = neo4j_rel.type
    rel_type_data = rel_classes[rel_type]

    if not rel_type_data:
        warnings.warn(
            (
                f"Could not find a class for {rel_type} relationship type."
                " Did you define the class before initializing Neontology?"
            )
        )
        return None

    if not neo4j_rel.start_node or not neo4j_rel.start_node.labels or not neo4j_rel.end_node or not neo4j_rel.end_node.labels:
        warnings.warn(
            (
                f"{rel_type} relationship type query did not include nodes."
                " To get neontology relationships, return source and target "
                "nodes as part of result."
            )
        )
        return None

    src_node = neo4j_node_to_neontology_node(neo4j_rel.start_node, node_classes)
    tgt_node = neo4j_node_to_neontology_node(neo4j_rel.end_node, node_classes)

    rel_props = convert_neo4j_types(dict(neo4j_rel))
    rel_props["source"] = src_node
    rel_props["target"] = tgt_node

    return rel_type_data.relationship_class(**rel_props)


def neo4j_records_to_neontology_records(
    records: list[Neo4jRecord],
    node_classes: dict,
    rel_classes: dict[str, "RelationshipTypeData"],
) -> tuple:
    """Convert native Neo4j records to Neontology records.

    Args:
        records (list[Neo4jRecord]): List of Neo4j records to convert.
        node_classes (dict): Mapping of labels to node classes used for populating with results.
        rel_classes (dict[str, RelationshipTypeData]): Mapping of relationship types to classes for populating with results.

    Returns:
        tuple: A tuple containing:
            - new_records (list): List of converted records in Neontology format.
            - unique_nodes (list): List of unique nodes.
            - rels (list): List of relationships.
            - paths (list): List of paths.
    """
    new_records = []

    for record in records:
        new_record: dict[str, dict] = {"nodes": {}, "relationships": {}, "paths": {}}

        for key, entry in record.items():
            if isinstance(entry, Neo4jNode):
                neontology_node = neo4j_node_to_neontology_node(entry, node_classes)

                if neontology_node:
                    new_record["nodes"][key] = neontology_node

            elif isinstance(entry, Neo4jRelationship):
                neontology_rel = neo4j_relationship_to_neontology_rel(entry, node_classes, rel_classes)

                if neontology_rel:
                    new_record["relationships"][key] = neontology_rel

            elif isinstance(entry, Neo4jPath):
                entry_path = []
                for step in entry:
                    step_rel = neo4j_relationship_to_neontology_rel(step, node_classes, rel_classes)
                    entry_path.append(step_rel)
                if entry_path:
                    new_record["paths"][key] = entry_path

        new_records.append(new_record)

    nodes_list_of_lists = [x["nodes"].values() for x in new_records]

    nodes = list(itertools.chain.from_iterable(nodes_list_of_lists))

    nodes_map = {f"{x.__primarylabel__}:{x.get_pp()}": x for x in nodes}

    unique_nodes = list(nodes_map.values())

    rels_list_of_lists = [x["relationships"].values() for x in new_records]
    rels = list(itertools.chain.from_iterable(rels_list_of_lists))

    paths_list_of_lists = [x["paths"].values() for x in new_records]
    paths = list(itertools.chain.from_iterable(paths_list_of_lists))

    return new_records, unique_nodes, rels, paths


class Neo4jEngine(GraphEngineBase):
    def __init__(self, config: "Neo4jConfig") -> None:
        """Initialise connection to the engine.

        Args:
            config (Neo4jConfig): Takes a Neo4jConfig object.
        """
        self.driver = GraphDatabase.driver(
            config.connection_uri,
            auth=(config.connection_username, config.connection_password),
        )

    def verify_connection(self) -> bool:
        """Verify the connection to the Neo4j database.

        Returns:
            bool: True if the connection is successful, False otherwise.
        """
        try:
            self.driver.verify_connectivity()
            return True
        except:  # noqa: E722
            return False

    def close_connection(self) -> None:
        """Close the connection to the Neo4j database."""
        try:
            self.driver.close()
        except AttributeError:
            pass

    def evaluate_query(
        self,
        cypher: LiteralString,
        params: dict = {},
        node_classes: dict = {},
        relationship_classes: dict = {},
    ) -> NeontologyResult:
        """Evaluate a Cypher query and return the results as Neontology records.

        Args:
            cypher (str): query to evaluate.
            params (dict, optional): parameters to pass through. Defaults to {}.
            node_classes (dict, optional): mapping of labels to node classes used for populating with results. Defaults to {}.
            relationship_classes (dict, optional): mapping of relationship types to classes used for populating with results.
                Defaults to {}.

        Returns:
            NeontologyResult: Result object containing the records, nodes, relationships, and paths.
        """
        result = self.driver.execute_query(cypher, parameters_=params)

        neo4j_records = result.records
        neontology_records, nodes, rels, paths = neo4j_records_to_neontology_records(
            neo4j_records, node_classes, relationship_classes
        )

        return NeontologyResult(
            records_raw=neo4j_records,
            records=neontology_records,
            nodes=nodes,
            relationships=rels,
            paths=paths,
        )

    def evaluate_query_single(self, cypher: LiteralString, params: dict = {}) -> Optional[Any]:
        """Evaluate a Cypher query which returns a single result.

        Args:
            cypher (str): query to evaluate.
            params (dict, optional): parameters to pass through. Defaults to {}.

        Returns:
            Optional[Any]: Query result, or None if no result is found.
        """
        result = self.driver.execute_query(cypher, parameters_=params, result_transformer_=Neo4jResult.single)

        if result:
            return result.value()

        else:
            return None

    def apply_constraint(self, label: str, property: str) -> None:
        """Apply a constraint to ensure that a property is unique for a label.

        Args:
            label (str): The label to which the constraint applies.
            property (str): The property that must be unique.
        """
        cypher = f"""
        CREATE CONSTRAINT IF NOT EXISTS
        FOR (n:{gql_identifier_adapter.validate_strings(label)})
        REQUIRE n.{gql_identifier_adapter.validate_strings(property)} IS UNIQUE
        """

        self.evaluate_query_single(cast(LiteralString, cypher))

    def drop_constraint(self, constraint_name: str) -> None:
        """Drop a constraint by its name.

        Args:
            constraint_name (str): The name of the constraint to drop.
        """
        drop_cypher = f"""
        DROP CONSTRAINT {gql_identifier_adapter.validate_strings(constraint_name)}
        """
        self.evaluate_query_single(cast(LiteralString, drop_cypher))

    def get_constraints(self) -> list:
        """Get the constraints defined in the graph.

        Returns:
            list: A list of constraint names.
        """
        get_constraints_query = """
        SHOW CONSTRAINTS yield name
        RETURN COLLECT(DISTINCT name)
        """

        constraints = self.evaluate_query_single(get_constraints_query)

        if constraints:
            return constraints

        else:
            return []


class Neo4jConfig(GraphEngineConfig):
    """Configuration for a Neo4j graph engine."""

    engine: ClassVar[type[GraphEngineBase]] = Neo4jEngine
    uri: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None

    env_fields: ClassVar[dict[str, str]] = {
        "uri": "NEO4J_URI",
        "username": "NEO4J_USERNAME",
        "password": "NEO4J_PASSWORD",
    }

    # Properties that guarantee non-None values for type checking
    @property
    def connection_uri(self) -> str:
        """Get the URI, guaranteed to be non-None after validation."""
        if self.uri is None:
            raise ValueError("URI should be set by validator")

        return self.uri

    @property
    def connection_username(self) -> str:
        """Get the username, guaranteed to be non-None after validation."""
        if self.username is None:
            raise ValueError("Username should be set by validator")

        return self.username

    @property
    def connection_password(self) -> str:
        """Get the password, guaranteed to be non-None after validation."""
        if self.password is None:
            raise ValueError("Password should be set by validator")

        return self.password



import itertools
import json
import warnings
from typing import Any, ClassVar, Optional, TypeVar

import pandas as pd
from pydantic import BaseModel, PrivateAttr, ValidationError, model_validator

from neontology.graphconnection import GraphConnection

from .basenode import BaseNode
from .commonmodel import CommonModel
from .gql import gql_identifier_adapter
from .schema_utils import RelationshipSchema, SchemaProperty, extract_type_mapping

R = TypeVar("R", bound="BaseRelationship")


class BaseRelationship(CommonModel):  # pyre-ignore[13]
    source: BaseNode
    target: BaseNode

    __relationshiptype__: ClassVar[Optional[str]] = None

    _merge_on: list[str] = PrivateAttr()  # what relationship properties should we merge on

    def __init__(self, **data: dict):
        super().__init__(**data)

        # we can define 'abstract' relationships which don't have a label
        # these are to provide common properties to be used by subclassed relationships
        # but shouldn't be put in the graph or even instantiated
        if self.__relationshiptype__ is None:
            raise NotImplementedError("Relationships to be used in the graph must define a relationship type.")

    @classmethod
    def _set_prop_usage(cls) -> None:
        """Set the properties that are used by Neontology for specific purposes.

        This method initializes the class attributes `_merge_on` based on the model's JSON schema.
        It retrieves properties that are marked for merging on, which are used to determine how relationships
        should be merged into the database.
        It also calls the superclass method to set other common property usages.
        """
        super()._set_prop_usage()
        cls._merge_on = cls._get_prop_usage("merge_on")

    @model_validator(mode="after")
    def validate_identifiers(self) -> "BaseRelationship":
        """Validate the relationship type identifier.

        This method checks that the relationship type identifier contains only alphanumeric characters and underscores,
        and that it begins with an alphabetic character.
        If the validation fails, it raises a warning.

        Returns:
            BaseRelationship: The instance of the relationship after validation.
        """
        try:
            gql_identifier_adapter.validate_strings(self.__relationshiptype__)
        except AttributeError:
            pass
        except ValidationError:
            warnings.warn(
                (
                    "Relationship type should contain only alphanumeric characters and underscores."
                    " It should begin with an alphabetic character."
                )
            )

        return self

    @classmethod
    def get_relationship_type(cls) -> Optional[str]:
        """Get the relationship type to use for creating and matching this relationship.

        If __relationship__ has been specified, use that.

        Otherwise use the class name in uppercase

        Returns:
            str: the string to use for creating and matching this relationship
        """
        return cls.__relationshiptype__

    def _get_merge_parameters(self, source_prop: str, target_prop: str) -> dict[str, Any]:
        """Get the parameters to use for merging this relationship.

        This method retrieves the properties of the source and target nodes,
        as well as the properties defined in the relationship itself that should be merged.
        It constructs a dictionary of parameters that can be used to merge the relationship
        into the database.

        Args:
            source_prop (str): The property of the source node to use for merging.
            target_prop (str): The property of the target node to use for merging.

        Returns:
            dict[str, Any]: A dictionary of parameters to use for merging the relationship.
        """
        exclusions = {"source", "target"}

        params = self._get_merge_parameters_common(exclude=exclusions)
        # get all the properties
        all_props = params.pop("all_props")

        # merge_props properties will be referenced individually with kwargs
        merge_props = {k: all_props[k] for k in self._merge_on}

        source_prop = self.source.model_dump()[source_prop]
        target_prop = self.target.model_dump()[target_prop]

        params.update(
            {
                "source_prop": source_prop,
                "target_prop": target_prop,
                **merge_props,
            }
        )

        return params

    def merge(
        self,
    ) -> None:
        """Merge this relationship into the database."""
        source_label = self.source.__primarylabel__
        target_label = self.target.__primarylabel__

        if not source_label or not target_label:
            raise ValueError("Source and target Nodes must have a defined primary label for creating a relationship.")

        source_prop = self.source.__primaryproperty__
        target_prop = self.target.__primaryproperty__

        rel_props = self._get_merge_parameters(source_prop=source_prop, target_prop=target_prop)

        merge_on_props = self._merge_on

        rel_type = self.get_relationship_type()

        if not rel_type:
            raise ValueError("Realtionship must have a defined relationship type for creating a relationship.")

        gc = GraphConnection()

        gc.merge_relationships(
            source_label,
            target_label,
            source_prop,
            target_prop,
            rel_type,
            merge_on_props,
            [rel_props],
        )

    @classmethod
    def merge_relationships(
        cls: type[R],
        rels: list[R],
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ) -> None:
        """Merge multiple relationships (of this type) into the database.

        Sometimes the source and target label may be ambiguous (e.g. where we have subclassed nodes)
            In this case you can explicitly pass in the relevant types

        Sometimes we want to match nodes on a property which isn't the primary property,
        so we can specify what property to use.

        Args:
            cls (type[R]): this class
            rels (list[R]): a list of relationships which are instances of this class
            source_prop (Optional[str]): explicitly specify the property to use for the source node
                if None, will use the primary property of the source node class
            target_prop (Optional[str]): explicitly specify the property to use for the target node
                if None, will use the primary property of the target node class

        """
        # define the properties to merge on
        merge_on_props = cls._get_prop_usage("merge_on")

        # sources and targets could have different primary labels
        # to operate efficiently, we group like source and targets for batch creation of relationships
        grouped_rels = itertools.groupby(rels, lambda x: (x.source.__class__, x.target.__class__))

        for node_clases, common_rels in grouped_rels:
            src_class = node_clases[0]
            tgt_class = node_clases[1]

            source_label = src_class.__primarylabel__
            target_label = tgt_class.__primarylabel__

            if source_prop is None:
                source_prop = src_class.__primaryproperty__

            if target_prop is None:
                target_prop = tgt_class.__primaryproperty__

            if not source_label or not target_label:
                raise ValueError("Source and target Nodes must have a defined primary label to create a relationship.")

            rel_type = cls.get_relationship_type()

            if not rel_type:
                raise ValueError("Relationship must have a defined relationship type for creating a relationship.")

            rel_list: list[dict[str, Any]] = [x._get_merge_parameters(source_prop, target_prop) for x in common_rels]

            gc = GraphConnection()

            gc.merge_relationships(
                source_label,
                target_label,
                source_prop,
                target_prop,
                rel_type,
                merge_on_props,
                rel_list,
            )

    @classmethod
    def merge_records(
        cls: type[R],
        records: list[dict[str, Any]],
        source_type: Optional[type[BaseNode]] = None,
        target_type: Optional[type[BaseNode]] = None,
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ) -> None:
        """Take a list of dictionaries and use them to merge in relationships in the graph.

        Sometimes, a relationship can accept nodes which subclass a particular node type.
            In these instances, it may be necessary to explicitly state what type of node should be used.

        Each record should have a source and target key where the value is the primary property
            value of the respective nodes.

        Args:
            records (list[dict[str, Any]]): a list of dictionaries used to populate relationships
            source_type: explicitly state the class to use for source node
            target_type: explicitly state the class to use for target node
            source_prop: explicitly state the property to use for the source node
            target_prop: explicitly state the property to use for the target node

        """
        hydrated_list = []

        if source_type is None:
            source_type = cls.model_fields["source"].annotation

        if target_type is None:
            target_type = cls.model_fields["target"].annotation

        if not source_type or not target_type:
            raise ValueError("Source and target Nodes types not defined.")

        if source_prop is None:
            source_prop = source_type.__primaryproperty__

        if target_prop is None:
            target_prop = target_type.__primaryproperty__

        for record in records:
            hydrated = dict(record)

            hydrated["source"] = source_type.model_construct(**{source_prop: record["source"]})
            hydrated["target"] = target_type.model_construct(**{target_prop: record["target"]})

            hydrated_list.append(cls(**hydrated))

        cls.merge_relationships(
            hydrated_list,
            source_prop=source_prop,
            target_prop=target_prop,
        )

    @classmethod
    def merge_df(
        cls: type[R],
        df: pd.DataFrame,
        source_type: Optional[type[BaseNode]] = None,
        target_type: Optional[type[BaseNode]] = None,
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ) -> None:
        """Merge in relationships based on data in a pandas data frame.

        Expects columns named 'source' and 'target' with the primary property value
            for the source and target nodes.

        Then additional fields should have a corresponding column.

        Args:
            df (pd.DataFrame): pandas dataframe where each row represents a relationship to merge.
            source_type (Optional[type[BaseNode]]): The class to use for the source node.
            target_type (Optional[type[BaseNode]]): The class to use for the target node.
            source_prop (Optional[str]): The property to use for the source node.
            target_prop (Optional[str]): The property to use for the target node.
        """
        if df.empty is False:
            cleaned_df = df.mask(pd.isna(df), None).copy()
            records = cleaned_df.to_dict(orient="records")
            cls.merge_records(
                records,
                source_type=source_type,
                source_prop=source_prop,
                target_type=target_type,
                target_prop=target_prop,
            )

    @classmethod
    def match_relationships(cls, limit: Optional[int] = None, skip: Optional[int] = None) -> list["BaseRelationship"]:
        """Match relationships of this type in the graph.

        Constructs a Cypher query to match relationships of the specified type in the graph database.
        It uses the GraphConnection class to execute the query and return a list of relationships.

        Args:
            cls (type[R]): The class of the relationship to match.
            limit (Optional[int]): The maximum number of relationships to return.
            skip (Optional[int]): The number of relationships to skip before returning results.

        Returns:
            list[BaseRelationship]: A list of relationships of the specified type.
        """
        gc = GraphConnection()
        result = gc.match_relationships(cls, limit, skip)

        return result

    @classmethod
    def get_count(cls):
        """Get the count of relationships of this type in the graph.

        Constructs a Cypher query to count the number of distinct relationships
            of the specified type in the graph database.
        It uses the GraphConnection class to execute the query and return the count.

        Returns:
            int: The count of distinct relationships of the specified type.
        """
        gc = GraphConnection()
        cypher = f"MATCH (n)-[r:{cls.__relationshiptype__}]->(o) RETURN COUNT(DISTINCT r)"
        result = gc.evaluate_query_single(cypher)
        return result

    def _prep_dump_dict(self, dumped_model: dict, exclude_node_props: bool = True) -> dict:
        """Prepare the dumped model dictionary for Neontology.

        This method modifies the dumped model dictionary to include additional metadata
        such as the source and target node labels and the relationship type.

        Args:
            dumped_model (dict): The dumped model dictionary.
            exclude_node_props (bool): Whether to exclude the source and target node properties.

        Returns:
            dict: The modified dumped model dictionary.
        """
        if exclude_node_props is True:
            dumped_model["source"] = self.source.get_pp()
            dumped_model["SOURCE_LABEL"] = self.source.__primarylabel__
            dumped_model["target"] = self.target.get_pp()
            dumped_model["TARGET_LABEL"] = self.target.__primarylabel__

        else:
            dumped_model["source"]["LABEL"] = self.source.__primarylabel__
            dumped_model["source"]["PK"] = self.source.get_pp()

            dumped_model["target"]["LABEL"] = self.target.__primarylabel__
            dumped_model["target"]["PK"] = self.target.get_pp()

        dumped_model["RELATIONSHIP_TYPE"] = self.__relationshiptype__

        return dumped_model

    def neontology_dump(
        self,
        exclude_node_props: bool = True,
        exclude: Optional[set] = None,
        exclude_none: bool = True,
        **kwargs,
    ) -> dict:
        """Dump the relationship as a dictionary.

        The generated dictionary can be used to create a relationship in Neontology.
        It includes additional metadata such as the source and target node labels and the relationship type.

        Args:
            exclude_node_props (bool): Whether to exclude the source and target node properties.
                Defaults to True.
            exclude (Optional[set]): A set of properties to exclude from the dump.
                Defaults to None.
            exclude_none (bool): Whether to exclude properties with None values.
                Defaults to True.
            **kwargs: Additional keyword arguments to pass to the model_dump method.

        Returns:
            dict: A dictionary representation of the relationship.
        """
        dumped_model = self.model_dump(exclude_none=exclude_none, exclude=exclude, **kwargs)

        return self._prep_dump_dict(dumped_model, exclude_node_props)

    def neontology_dump_json(
        self,
        exclude_node_props: bool = True,
        exclude: Optional[set] = None,
        exclude_none: bool = True,
        **kwargs,
    ) -> str:
        """Dump the relationship as a JSON string.

        The generated JSON can be imported with Neontology. It includes additional metadata
        such as the source and target node labels and the relationship type.

        Args:
            exclude_node_props (bool): Whether to exclude the source and target node properties.
                Defaults to True.
            exclude (Optional[set]): A set of properties to exclude from the dump.
                Defaults to None.
            exclude_none (bool): Whether to exclude properties with None values.
                Defaults to True.
            **kwargs: Additional keyword arguments to pass to the model_dump method.

        Returns:
            str: A JSON string representation of the relationship.
        """
        # pydantic converts values to be json serializable, make use of this first
        original_json = self.model_dump_json(exclude_none=exclude_none, exclude=exclude, **kwargs)
        model_dict = json.loads(original_json)

        return json.dumps(self._prep_dump_dict(model_dict, exclude_node_props))

    @classmethod
    def neontology_schema(
        cls,
        source_labels: Optional[list[str]] = None,
        target_labels: Optional[list[str]] = None,
    ) -> RelationshipSchema:
        """Generate a schema for this relationship type.

        Args:
            source_labels (Optional[list[str]]): Labels for the source node type.
            target_labels (Optional[list[str]]): Labels for the target node type.

        Returns:
            RelationshipSchema: A schema object representing the relationship type.
        """
        schema_properties: list[SchemaProperty] = []
        rel_type = cls.__relationshiptype__

        if not rel_type:
            raise ValueError("Relationship doesn't have a relationship type.")

        for field_name, field_props in cls.model_fields.items():
            if field_name in ["source", "target"]:
                continue

            field_type = extract_type_mapping(field_props.annotation, show_optional=True)

            required_field = field_props.is_required()

            rel_prop = SchemaProperty(
                type_annotation=field_type,
                name=field_name,
                required=required_field,
            )

            if required_field is True:
                schema_properties.insert(0, rel_prop)

            else:
                schema_properties.append(rel_prop)

        source_type = cls.model_fields["source"].annotation
        target_type = cls.model_fields["target"].annotation

        if not target_labels:
            if getattr(target_type, "__primarylabel__", None):
                target_labels = [target_type.__primarylabel__]

        if not source_labels:
            if getattr(source_type, "__primarylabel__", None):
                source_labels = [source_type.__primarylabel__]

        if not source_labels or not target_labels:
            from neontology.utils import get_node_types

            if not target_labels:
                # return concrete subclasses of the abstract node class given
                retrieved_node_types = get_node_types(target_type)
                target_labels = list(retrieved_node_types.keys())

            if not source_labels:
                # return concrete subclasses of the abstract node class given
                retrieved_node_types = get_node_types(source_type)
                source_labels = list(retrieved_node_types.keys())

        schema = RelationshipSchema(
            name=rel_type,
            relationship_type=rel_type,
            properties=schema_properties,
            target_labels=target_labels,
            source_labels=source_labels,
        )

        return schema


class RelationshipTypeData(BaseModel):
    relationship_class: type[BaseRelationship]
    source_class: type[BaseNode]
    target_class: type[BaseNode]
    all_source_classes: list[type[BaseNode]]
    all_target_classes: list[type[BaseNode]]



import json
from hashlib import sha1
from typing import Any

from pydantic import BaseModel, computed_field


class NeontologyResult(BaseModel):
    records_raw: Any
    records: list
    nodes: list
    relationships: list
    paths: list

    @computed_field  # type: ignore[misc]
    @property
    def node_link_data(self) -> dict:
        """Get the result as a dictionary with 'nodes' and 'edges' keys.

        Returns:
            dict: Dictionary with 'nodes' and 'edges' keys, suitable for use with
            networkx or network visualisation libraries like D3.js or Cytoscape.js.
        """
        nodes = {
            f"{x.__primarylabel__}:{str(x.get_pp())}": {
                **x.neontology_dump(),
                **{
                    "__pp__": x.get_pp(),
                    "__str__": str(x),
                },
            }
            for x in self.nodes
        }

        links = {sha1(x.neontology_dump_json().encode("utf-8")).hexdigest(): x.neontology_dump() for x in self.relationships}

        # deduplicate nodes and links
        unique_nodes = list(nodes.values())
        unique_links = list(links.values())

        data = {"nodes": unique_nodes, "edges": unique_links, "directed": True}

        return data

    def neontology_dump(self) -> dict:
        """Dump the results as a dictionary, in 'Neontology' format.

        This includes LABEL and RELATIONSHIP_TYPE keys for easy import with Neontology.
        """
        nodes = [x.neontology_dump() for x in self.nodes]
        relationships = [x.neontology_dump() for x in self.relationships]

        data = {"nodes": nodes, "edges": relationships}

        return data

    def neontology_dump_json(self) -> str:
        """Dump the results as a JSON string, in 'Neontology' format.

        This includes LABEL and RELATIONSHIP_TYPE keys for easy import with Neontology.
        """
        nodes = [json.loads(x.neontology_dump_json()) for x in self.nodes]
        relationships = [json.loads(x.neontology_dump_json()) for x in self.relationships]

        data = {"nodes": nodes, "edges": relationships}

        return json.dumps(data)



import logging
from collections import defaultdict
from copy import deepcopy
from typing import Any, Optional, Union

from pydantic import BaseModel, model_validator

from ..gql import gql_identifier_adapter
from ..graphconnection import GraphConnection
from ..utils import get_node_types, get_rels_by_type

logger = logging.getLogger(__name__)


def _nested_dict(n, type):
    """Create a nested defaultdict of depth n with the given type."""
    if n == 1:
        return defaultdict(type)
    else:
        return defaultdict(lambda: _nested_dict(n - 1, type))


class NeontologyNodeRaw(BaseModel):
    """Pydantic model for a Neontology node raw record."""

    LABEL: str


class NeontologyRelationshipRaw(BaseModel):
    """Pydantic model for a Neontology relationship raw record."""

    RELATIONSHIP_TYPE: str
    TARGET_LABEL: str
    TARGET_NODES: Optional[list[NeontologyNodeRaw]] = None
    TARGETS: Optional[list[str]] = None
    TARGET_PROPERTY: Optional[str] = None


class NeontologyNodeRecord(BaseModel):
    """Pydantic model for transforming a raw Neontology node record."""

    input_record: NeontologyNodeRaw
    label: Optional[str] = None
    output_record: Optional[dict[str, Any]] = None

    @model_validator(mode="before")
    def populate_node_fields(cls, data):
        """Use the raw input data (from input_record) to populate the fields of the model."""
        if not data.get("label"):
            data["label"] = data.get("input_record", {}).get("LABEL")

        if not data.get("output_record"):
            data["output_record"] = {k: v for k, v in data.get("input_record", {}).items() if k not in ["LABEL"]}

        return data


class NeontologyRelationshipRecord(BaseModel):
    """Pydantic model for transforming a raw Neontology relationship record."""

    input_record: NeontologyRelationshipRaw
    relationship_type: Optional[str] = None

    target_prop: Optional[str] = None
    source_label: Optional[str] = None
    target_label: Optional[str] = None
    source: Optional[Any] = None
    target: Optional[Any] = None

    relationship_properties: Optional[dict] = None
    output_record: Optional[dict] = None

    @model_validator(mode="before")
    def populate_rel_fields(cls, data: dict) -> dict:
        """Use the raw input data (from input_record) to populate the fields of the model.

        Args:
            data: dict of raw model fields before parsing.

        Returns:
            dict: processed dict of model fields ready for Pydantic validation.
        """
        if not data.get("RELATIONSHIP_TYPE"):
            data["relationship_type"] = data.get("input_record", {}).get("RELATIONSHIP_TYPE")

        if not data.get("source_label"):
            data["source_label"] = data.get("input_record", {}).get("SOURCE_LABEL")

        if not data.get("target_label"):
            data["target_label"] = data.get("input_record", {}).get("TARGET_LABEL")

        if not data.get("target_prop"):
            data["target_prop"] = data.get("input_record", {}).get("TARGET_PROPERTY")

        if not data.get("source"):
            data["source"] = data.get("input_record", {}).get("source")

        if not data.get("target"):
            data["target"] = data.get("input_record", {}).get("target")

        if not data.get("relationship_properties"):
            data["relationship_properties"] = {
                k: v
                for k, v in data.get("input_record", {}).items()
                if k
                not in [
                    "SOURCE_LABEL",
                    "TARGET_LABEL",
                    "TARGET_PROPERTY",
                    "RELATIONSHIP_TYPE",
                    "TARGET_NODES",
                    "TARGETS",
                    "source",
                    "target",
                ]
            }

        if not data.get("output_record"):
            data["output_record"] = {
                k: v
                for k, v in data.get("input_record", {}).items()
                if k
                not in [
                    "SOURCE_LABEL",
                    "TARGET_LABEL",
                    "TARGET_PROPERTY",
                    "RELATIONSHIP_TYPE",
                    "TARGET_NODES",
                    "TARGETS",
                ]
            }

        return data

    @model_validator(mode="after")
    def populate_fields(self):
        """Use the raw input data (from input_record) to populate the fields of the model."""
        node_types = get_node_types()

        # if an explicit TARGET_PROPERTY wasn't passed in, use the primary property
        if self.target_prop is None:
            self.target_prop = node_types[self.target_label].__primaryproperty__

        return self


def _import_nodes(input_records: list[NeontologyNodeRecord]):
    mapped_records = defaultdict(list)
    node_types = get_node_types()

    for record in input_records:
        mapped_records[record.label].append(record.output_record)

    for label, node_records in mapped_records.items():
        node_class = node_types[label]
        node_class.merge_records(node_records)


def _import_relationships(
    input_records: list[NeontologyRelationshipRecord],
    check_unmatched: bool,
    error_on_unmatched: bool,
):
    # if we're meant to warn on unmatched, we need to look up source and target

    mapped_records = _nested_dict(4, list)
    rel_types = get_rels_by_type()
    node_types = get_node_types()

    for record in input_records:
        rel_type = record.relationship_type

        # when we merge relationship records, we pass in target prop and source prop
        # we also need to hydrate based on the given source and target labels
        # therefore we need to group together records which share those properties
        mapped_records[rel_type][record.target_prop][record.source_label][record.target_label].append(record)

    for rel_type, rel_records_by_prop in mapped_records.items():
        for target_prop, rel_records_by_source_label in rel_records_by_prop.items():
            for (
                source_label,
                rel_records_by_target_label,
            ) in rel_records_by_source_label.items():
                for target_label, rel_entries in rel_records_by_target_label.items():
                    if check_unmatched is True:
                        for rel_entry in rel_entries:
                            gc = GraphConnection()

                            gql_identifier_adapter.validate_strings(target_label)
                            gql_identifier_adapter.validate_strings(target_prop)

                            cypher = f"""
                                MATCH (n:{target_label})
                                WHERE n.{target_prop} = $val
                                RETURN n
                                """

                            params = {"val": rel_entry.target}

                            result = gc.evaluate_query(cypher, params)

                            if len(result.nodes) > 1:
                                message = (
                                    f"Matched {len(result.nodes)} on {rel_type} for {target_label}"
                                    f" WHERE {target_prop} = {rel_entry.target}"
                                )

                                if error_on_unmatched is True:
                                    raise ValueError(message)

                                logger.warning(message)

                            elif len(result.nodes) == 0:
                                error_msg = (
                                    f"No target node for {rel_type} to {rel_entry.target} "
                                    "Ensure primary properties are explicitly set or deterministic."
                                )

                                if error_on_unmatched is True:
                                    raise ValueError(error_msg)

                                logger.warning(error_msg)

                    source_type = node_types[source_label]
                    target_type = node_types[target_label]

                    rel_class = rel_types[rel_type].relationship_class

                    output_records = [x.output_record for x in rel_entries]

                    rel_class.merge_records(
                        output_records,
                        source_type=source_type,
                        target_type=target_type,
                        target_prop=target_prop,
                    )


def _process_sub_records(
    source_node_record: NeontologyNodeRecord, subrecords
) -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
    node_types = get_node_types()

    output_nodes = []
    output_rels = []

    # hydrate the source node to get the pp
    # then dump it

    source_label = source_node_record.label

    if source_label is None:
        raise ValueError("source_node_record.label is None, cannot look up node type.")

    source_class = node_types[source_label]

    if not source_node_record.output_record:
        raise TypeError("Expected output dictionary.")

    source_node = source_class(**source_node_record.output_record)
    source_node_pk = source_node.get_pp()

    # iterate through the subrecords

    for record in subrecords:
        target_prop = record.get("TARGET_PROPERTY")
        target_class = node_types[record["TARGET_LABEL"]]

        rel_targets = []

        # handle new nodes that need to be created
        # we need to hydrate these to get the pp
        # then dump them
        if "TARGET_NODES" in record:
            for entry in record["TARGET_NODES"]:
                target_node_record = NeontologyNodeRecord(input_record=entry)
                tgt_node = target_class(**target_node_record.output_record)
                rel_targets.append(tgt_node.get_pp())
                output_nodes.append(target_node_record)

        if "TARGETS" in record:
            rel_targets += record["TARGETS"]

        # now do the relationships
        # convert to neontology style
        rel_dict = record
        rel_dict["source"] = source_node_pk
        rel_dict["SOURCE_LABEL"] = source_label

        if target_prop:
            rel_dict["TARGET_PROPERTY"] = target_prop

        output_raw_records = [{**rel_dict, **{"target": x}} for x in rel_targets]

        output_rels += [NeontologyRelationshipRecord(input_record=x) for x in output_raw_records]

    # returns a set of records to be used as input to import_records
    return output_nodes, output_rels


def _prepare_records(
    input_records: Union[list[dict[str, Any]], dict[str, Any]],
) -> tuple:
    input_records = input_records.copy()

    if isinstance(input_records, dict):
        # handle the situation where we've just been passed a single record
        if "nodes" not in input_records and "edges" not in input_records:
            raw_records = [input_records]

        # handle the situation where we've got 'link data'
        # which consists of node records and link/relationship records
        else:
            raw_records = input_records.get("nodes", [])
            raw_records += input_records.get("edges", [])

    else:
        raw_records = input_records.copy()

    input_nodes = []
    input_relationships = []

    for record in raw_records:
        if "LABEL" in record.keys():
            # pull out relationships
            # append the input_node
            # now parse the relationships
            rel_records = record.pop("RELATIONSHIPS_OUT", None)

            node_record = NeontologyNodeRecord(input_record=record)

            input_nodes.append(node_record)

            if rel_records:
                new_nodes, new_rels = _process_sub_records(node_record, rel_records)
                input_nodes += new_nodes
                input_relationships += new_rels

        elif "RELATIONSHIP_TYPE" in record.keys():
            input_relationships.append(NeontologyRelationshipRecord(input_record=record))

        else:
            raise ValueError(
                (
                    "Input record does not have LABEL (for node) or RELATIONSHIP_TYPE (for relationship). %s",
                    record,
                )
            )

    return input_nodes, input_relationships


def import_records(
    records: list,
    validate_only: bool = False,
    check_unmatched: bool = True,
    error_on_unmatched: bool = False,
) -> None:
    """Import 'Neontology' records into the graph database.

    Args:
        records (list): a list of dictionaries representing nodes and relationships.
        validate_only (bool, optional): validate the data is appropriately formatted
            but do not populate the graph. Defaults to False.
        check_unmatched (bool, optional): Check relationship targets exist, warn if not. Defaults to True.
        error_on_unmatched (bool, optional): When checking for relationship targets,
            raise an error if the target node is not in the graph. Defaults to False.
    """
    input_nodes = []
    input_rels = []

    fresh_records = deepcopy(records)

    for entry in fresh_records:
        prepared_nodes, prepared_rels = _prepare_records(entry)

        input_nodes += prepared_nodes
        input_rels += prepared_rels

    if validate_only is False:
        _import_nodes(input_nodes)
        _import_relationships(input_rels, check_unmatched, error_on_unmatched)



import json
from datetime import datetime
from enum import Enum
from typing import Annotated, ClassVar, Optional
from uuid import UUID, uuid4

import pandas as pd
import pytest
from pydantic import (
    ConfigDict,
    Field,
    ValidationInfo,
    field_serializer,
    field_validator,
)

from neontology import (
    BaseNode,
    BaseRelationship,
    GQLIdentifier,
    related_nodes,
    related_property,
)
from neontology.result import NeontologyResult


class PracticeNode(BaseNode):
    __primaryproperty__: ClassVar[GQLIdentifier] = "pp"
    __primarylabel__: ClassVar[Optional[GQLIdentifier]] = "PracticeNode"

    pp: str


class PracticeNodeDated(BaseNode):
    __primaryproperty__: ClassVar[GQLIdentifier] = "pp"
    __primarylabel__: ClassVar[Optional[GQLIdentifier]] = "PracticeNodeDated"

    pp: str

    test_merged: datetime = Field(
        default_factory=datetime.now,
    )

    # created property will only be set 'on create' - when the node is first created
    test_created: Optional[datetime] = Field(
        default=None, validate_default=True, json_schema_extra={"set_on_create": True}
    )

    @field_validator("test_created")
    def set_test_created_to_merged(
        cls, value: Optional[datetime], values: ValidationInfo
    ) -> datetime:
        """When the node is first created, we want the created value
        to be set equal to merged.
        Otherwise they will be a tiny amount of time different.
        """
        # if the created value has been manually set, don't override it
        if value is None:
            return values.data["test_merged"]
        else:
            return value


def test_set_pp_field_valid():
    tn = PracticeNode(pp="Some Value")

    assert tn.pp == "Some Value"


def test_engine_dict_internals():
    # verify that an error is raised when we
    # execute a function which depends on the db
    with pytest.raises(RuntimeError):
        PracticeNode.match_nodes()

    tn = PracticeNode(pp="Some Value")

    # should be able to call these functions without an engine connection
    assert tn._engine_dict() == {"pp": "Some Value"}


def test_get_merge_parameters():
    tn = PracticeNode(pp="Some Value")

    assert tn._get_merge_parameters() == {
        "always_set": {"pp": "Some Value"},
        "pp": "Some Value",
        "set_on_create": {},
        "set_on_match": {},
    }
    assert tn.get_pp() == "Some Value"


def test_create(use_graph):
    tn = PracticeNode(pp="Test Node")

    tn.create()

    cypher = """
    MATCH (n:PracticeNode)
    WHERE n.pp = 'Test Node'
    RETURN n
    """

    result = use_graph.evaluate_query(cypher)

    assert result.nodes[0].__primarylabel__ == "PracticeNode"

    assert result.nodes[0].pp == "Test Node"


def test_create_if_exists(use_graph):
    """Neontology does not check if a node already exists, it is for the user to enforce this at the database level."""
    tn = PracticeNode(pp="Test Node")

    tn.create()

    cypher = """
    MATCH (n:PracticeNode)
    WHERE n.pp = 'Test Node'
    RETURN n
    """

    result = use_graph.evaluate_query(cypher)

    assert result.nodes[0].__primarylabel__ == "PracticeNode"

    assert result.nodes[0].pp == "Test Node"

    tn.create()

    node_count = use_graph.evaluate_query_single(
        "MATCH (n:PracticeNode) WHERE n.pp = 'Test Node' RETURN COUNT(n)"
    )

    assert node_count == 2


def test_create_multiple_if_exists(use_graph):
    tn = PracticeNode(pp="Test Node")

    tn.create()

    cypher = """
    MATCH (n:PracticeNode)
    WHERE n.pp = 'Test Node'
    RETURN n
    """

    result = use_graph.evaluate_query(cypher)

    assert result.nodes[0].__primarylabel__ == "PracticeNode"

    assert result.nodes[0].pp == "Test Node"

    PracticeNode.create_nodes([tn])

    node_count = use_graph.evaluate_query_single(
        "MATCH (n:PracticeNode) WHERE n.pp = 'Test Node' RETURN COUNT(n)"
    )

    assert node_count == 2


def test_no_primary_label():
    class SpecialPracticeNode(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        pp: str

    with pytest.raises(AttributeError):
        SpecialPracticeNode(pp="Test Node")


def test_none_primary_label():
    class SpecialPracticeNode(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = None
        pp: str

    with pytest.warns(
        UserWarning,
        match="Primary Label should contain only alphanumeric characters and underscores",
    ):
        with pytest.raises(NotImplementedError):
            SpecialPracticeNode(pp="Test Node")


def test_create_multilabel(use_graph):
    class MultipleLabelNode(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "PrimaryLabel"
        __secondarylabels__: ClassVar[Optional[list]] = ["ExtraLabel1", "ExtraLabel2"]
        pp: str

    tn = MultipleLabelNode(pp="Test Node")

    tn.create()

    cypher = """
    MATCH (n:ExtraLabel1)
    WHERE n.pp = 'Test Node'
    RETURN n
    """

    result = use_graph.evaluate_query(cypher)

    assert result.nodes[0].__primarylabel__ == "PrimaryLabel"

    # confirm the secondary labels were written to the database
    assert "ExtraLabel1" in result.records_raw[0].values()[0].labels
    assert "ExtraLabel2" in result.records_raw[0].values()[0].labels

    assert result.nodes[0].pp == "Test Node"


def test_create_multilabel_inheritance(use_graph):
    class Mammal(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __secondarylabels__: ClassVar[Optional[list]] = ["Mammal"]
        pp: str

    class Human(Mammal):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "Human"
        pp: str

    tn = Human(pp="Bob")

    tn.create()

    cypher = """
    MATCH (n:Human)
    WHERE n.pp = 'Bob'
    RETURN n
    """

    result = use_graph.evaluate_query(cypher)

    assert "Human" in result.records_raw[0].values()[0].labels
    assert "Mammal" in result.records_raw[0].values()[0].labels

    assert result.nodes[0].pp == "Bob"


def test_create_multilabel_inheritance_multiple(use_graph):
    class Animal(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __secondarylabels__: ClassVar[Optional[list]] = ["Animal"]
        pp: str

    class Human(Animal):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "Human"
        pp: str

    class Elephant(Animal):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "Elephant"
        pp: str

    human = Human(pp="Bob")

    human.create()

    elephant = Elephant(pp="Bob")

    # this should create a new node because the primary label is different
    elephant.create()

    elephant2 = Elephant(pp="Bob")

    # this shouldn't create a new node because the primary label is the same
    elephant2.merge()

    cypher = """
    MATCH (n:Animal)
    WHERE n.pp = 'Bob'
    RETURN COUNT(DISTINCT n)
    """

    result = use_graph.evaluate_query_single(cypher)

    assert result == 2


def test_merge_defined_label_inherited(use_graph):
    class Mammal(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __secondarylabels__: ClassVar[Optional[list]] = ["Mammal"]
        pp: str

    class Human(Mammal):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "Human"
        pp: str

    tn = Human(pp="Bob")

    tn.merge()

    cypher = """
    MATCH (n:Human)
    WHERE n.pp = 'Bob'
    RETURN n
    """

    result = use_graph.evaluate_query(cypher)

    assert "Human" in result.records_raw[0].values()[0].labels
    assert "Mammal" in result.records_raw[0].values()[0].labels

    assert result.nodes[0].pp == "Bob"


class SpecialPracticeNode(BaseNode):
    __primarylabel__: ClassVar[Optional[str]] = "SpecialTestLabel1"
    __primaryproperty__: ClassVar[str] = "pp"

    pp: str


def test_merge_multiple_defined_label(use_graph):
    tn = SpecialPracticeNode(pp="Special Test Node")

    tn2 = SpecialPracticeNode(pp="Special Test Node2")

    SpecialPracticeNode.merge_nodes([tn, tn2])

    cypher = """
    MATCH (n:SpecialTestLabel1)
    RETURN COLLECT(DISTINCT n.pp) as node_names
    """

    results = use_graph.evaluate_query_single(cypher)

    assert "Special Test Node" in results
    assert "Special Test Node2" in results


def test_create_multiple_defined_label(use_graph):
    tn = SpecialPracticeNode(pp="Special Test Node")

    tn2 = SpecialPracticeNode(pp="Special Test Node2")

    SpecialPracticeNode.create_nodes([tn, tn2])

    cypher = """
    MATCH (n:SpecialTestLabel1)
    RETURN COLLECT(DISTINCT n.pp) as node_names
    """

    results = use_graph.evaluate_query_single(cypher)

    assert "Special Test Node" in results
    assert "Special Test Node2" in results


def test_creation_datetime(use_graph):
    """Check we can manually define the created datetime, and then check we can
    query for it using neo4j DateTime type.
    """
    my_datetime = datetime(year=2022, month=5, day=4, hour=3, minute=21)

    bn = PracticeNodeDated(pp="Test Node", test_created=my_datetime)

    bn.create()

    cypher = """
    MATCH (n:PracticeNodeDated)
    WHERE n.pp = 'Test Node'
    RETURN n.test_created.year
    """

    result = use_graph.evaluate_query_single(cypher)

    assert result == 2022


def test_match_nodes(use_graph):
    tn = PracticeNode(pp="Special Test Node")

    tn2 = PracticeNode(pp="Special Test Node2")

    PracticeNode.merge_nodes([tn, tn2])

    results = PracticeNode.match_nodes()

    for result in results:
        assert isinstance(result, PracticeNode)

    pps = [x.pp for x in results]

    assert "Special Test Node" in pps

    assert "Special Test Node2" in pps


def test_match_nodes_limit(use_graph):
    tn = PracticeNode(pp="Special Test Node")
    tn.merge()

    tn2 = PracticeNode(pp="Special Test Node2")

    tn2.merge()

    results = PracticeNode.match_nodes(limit=1)

    assert len(results) == 1

    assert results[0].pp in ["Special Test Node", "Special Test Node2"]


def test_match_nodes_skip(use_graph):
    tn = PracticeNode(pp="Special Test Node")
    tn.merge()

    tn2 = PracticeNode(pp="Special Test Node2")

    tn2.merge()

    results1 = PracticeNode.match_nodes(limit=1)

    assert len(results1) == 1

    assert results1[0].pp in ["Special Test Node", "Special Test Node2"]

    results2 = PracticeNode.match_nodes(limit=1, skip=1)

    assert len(results2) == 1

    # make sure we have actually skipped through the results
    assert results1[0].pp != results2[0].pp


def test_match_node(use_graph):
    tn = PracticeNode(pp="Special Test Node")

    tn.create()

    result = PracticeNode.match("Special Test Node")

    assert isinstance(result, PracticeNode)

    assert "Special Test Node" == result.pp


def test_match_none(use_graph):
    bn = PracticeNode(pp="Test Node")
    bn.create()

    result = PracticeNode.match("Does Not Exist")

    assert result is None


def test_match_many_none(use_graph):
    class PracticeNode2(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[str] = "PracticeNode2"

        pp: str

    bn = PracticeNode2(pp="Test Node")
    bn.create()

    results = PracticeNode.match_nodes()

    assert results == []


def test_delete_node(use_graph):
    tn = PracticeNode(pp="Special Test Node")

    tn.create()

    result = PracticeNode.match("Special Test Node")

    assert isinstance(result, PracticeNode)

    PracticeNode.delete("Special Test Node")

    result = PracticeNode.match("Special Test Node")

    assert result is None


class ModelTestString(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "TestModelString"
    pp: str
    test_prop_string: str


class ModelTestInt(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "TestModelInt"
    pp: str
    test_prop_int: int


class ModelTestTuple(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "TestModelTuple"
    pp: str
    test_prop_tuple: tuple


class ModelTestSet(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "TestModelSet"
    pp: str
    test_prop_set: set


class ModelTestUUID(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "TestModelUUID"
    pp: str
    test_prop_uuid: UUID


class ModelTestDateTime(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "TestModelDateTime"
    pp: str
    test_prop_datetime: datetime


class ModelTestStringList(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "TestModelStringList"
    pp: str
    test_prop_list: list


class ModelTestIntListExplicit(BaseNode):
    __primaryproperty__: ClassVar[str] = "pp"
    __primarylabel__: ClassVar[Optional[str]] = "TestModelIntListExplicit"
    pp: str
    test_prop_int_list_exp: list[int]


@pytest.mark.parametrize(
    "test_model,test_prop,input_value,expected_value",
    [
        (ModelTestString, "test_prop_string", "hello world", ["hello world"]),
        (ModelTestInt, "test_prop_int", 5071, [5071]),
        (ModelTestTuple, "test_prop_tuple", ("hello", "world"), [["hello", "world"]]),
        (
            ModelTestSet,
            "test_prop_set",
            {"foo", "bar"},
            [["bar", "foo"], ["foo", "bar"]],
        ),
        (
            ModelTestUUID,
            "test_prop_uuid",
            UUID("32d4a4cb-29c3-4aa8-9b55-7790431819e3"),
            ["32d4a4cb-29c3-4aa8-9b55-7790431819e3"],
        ),
        (
            ModelTestDateTime,
            "test_prop_datetime",
            datetime(year=1984, month=1, day=2),
            [datetime(year=1984, month=1, day=2)],
        ),
        (
            ModelTestStringList,
            "test_prop_list",
            ["foo", "bar"],
            [["foo", "bar"]],
        ),
        (
            ModelTestIntListExplicit,
            "test_prop_int_list_exp",
            [1, 2, 3],
            [[1, 2, 3]],
        ),
    ],
)
def test_property_types(use_graph, test_model, test_prop, input_value, expected_value):
    pp = "test_node"

    input_data = {"pp": pp}
    input_data[test_prop] = input_value

    testmodel = test_model(**input_data)

    testmodel.create()

    result = test_model.match(pp)

    assert result.model_dump()[test_prop] == input_value

    cypher = f"""
    MATCH (n:{test_model.__primarylabel__})
    WHERE n.pp = 'test_node'
    RETURN n.{test_prop}
    """

    cypher_result = use_graph.evaluate_query_single(cypher)

    # in the case of sets, we may get the result back ordered one of two ways
    # therefore, we check that the result is one of the expected values rather
    assert cypher_result in expected_value


def test_empty_list_property(use_graph):
    class TestModelListProp(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "TestModel1"
        pp: str
        list_prop: list

    pp = "test_node"

    testmodel = TestModelListProp(list_prop=[], pp=pp)

    testmodel.create()

    result = TestModelListProp.match(pp)

    assert result.list_prop == []


def test_set_on_match(use_graph):
    """Check that we successfully identify field to set on match"""

    class TestModel(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "TestModel2"
        pp: str = "test_node"
        only_set_on_match: Optional[str] = Field(
            json_schema_extra={"set_on_match": True}, default=None
        )
        normal_field: str

    test_node = TestModel(only_set_on_match="Foo", normal_field="Bar", pp="test_node")
    test_node.merge()

    cypher = """
    MATCH (n:TestModel2)
    WHERE n.pp = 'test_node'
    RETURN n
    """

    cypher_result = use_graph.evaluate_query(cypher)

    assert cypher_result.nodes[0].only_set_on_match is None
    assert cypher_result.nodes[0].normal_field == "Bar"

    test_node2 = TestModel(only_set_on_match="Foo", normal_field="Bar", pp="test_node")
    test_node2.merge()

    cypher_result2 = use_graph.evaluate_query(cypher)

    assert cypher_result2.nodes[0].only_set_on_match == "Foo"


def test_set_on_create(use_graph):
    """Check that we successfully identify field to set on match"""

    class TestModel(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        __primarylabel__: ClassVar[Optional[str]] = "TestModel3"
        pp: str = "test_node"
        only_set_on_create: str = Field(json_schema_extra={"set_on_create": True})
        normal_field: str

    test_node = TestModel(only_set_on_create="Foo", normal_field="Bar", pp="test_node")
    test_node.merge()

    cypher = """
    MATCH (n:TestModel3)
    WHERE n.pp = 'test_node'
    RETURN n
    """

    cypher_result = use_graph.evaluate_query(cypher)

    assert cypher_result.nodes[0].only_set_on_create == "Foo"
    assert cypher_result.nodes[0].normal_field == "Bar"

    test_node2 = TestModel(only_set_on_create="Fee", normal_field="Fi", pp="test_node")
    test_node2.merge()

    cypher_result2 = use_graph.evaluate_query(cypher)

    assert cypher_result2.nodes[0].only_set_on_create == "Foo"
    assert cypher_result2.nodes[0].normal_field == "Fi"


class Person(BaseNode):
    __primaryproperty__: ClassVar[str] = "identifier"
    __primarylabel__: ClassVar[str] = (
        "PersonLabel1"  # optionally specify the label to use
    )

    name: str
    age: int
    identifier: Optional[str] = Field(default=None, validate_default=True)

    @field_validator("identifier")
    def set_identifier(cls, v, values):
        if v is None:
            v = f"{values.data['name']}_{values.data['age']}"

        return v


def test_merge_df_with_duplicates(use_graph):
    people_records = [
        {"name": "arthur", "age": 70},
        {"name": "betty", "age": 65},
        {"name": "betty", "age": 65},
        {"name": "ted", "age": 50},
        {"name": "betty", "age": 75},
        {"name": "arthur", "age": 70},
    ]

    people_df = pd.DataFrame.from_records(people_records)

    results = Person.merge_df(people_df)

    names = [x.name for x in results]

    assert names == ["arthur", "betty", "betty", "ted", "betty", "arthur"]

    assert results[0].identifier == results[5].identifier

    assert results[1].identifier == results[2].identifier

    assert results[1].identifier != results[4].identifier

    assert results[3].name == "ted"


class Person2(BaseNode):
    __primaryproperty__: ClassVar[str] = "name"
    __primarylabel__: ClassVar[str] = (
        "PersonLabel2"  # optionally specify the label to use
    )

    name: str
    age: int
    favorite_colors: Optional[list] = None


def test_merge_df_with_lists(use_graph):
    people_records = [
        {"name": "arthur", "age": 70, "favorite_colors": ["red"]},
        {"name": "betty", "age": 65, "favorite_colors": ["red", "blue"]},
        {"name": "ted", "age": 50, "favorite_colors": []},
        {"name": "ben", "age": 75},
    ]

    people_df = pd.DataFrame.from_records(people_records)

    Person2.merge_df(people_df, deduplicate=False)

    arthur = Person2.match("arthur")
    assert arthur.favorite_colors == ["red"]

    betty = Person2.match("betty")
    assert betty.favorite_colors == ["red", "blue"]

    ted = Person2.match("ted")
    assert ted.favorite_colors == []

    ben = Person2.match("ben")
    assert ben.favorite_colors is None


def test_get_count(use_graph):
    people_records = [
        {"name": "arthur", "age": 70, "favorite_colors": ["red"]},
        {"name": "betty", "age": 65, "favorite_colors": ["red", "blue"]},
        {"name": "ted", "age": 50, "favorite_colors": []},
        {"name": "ben", "age": 75},
    ]

    people_df = pd.DataFrame.from_records(people_records)

    Person2.merge_df(people_df, deduplicate=False)

    assert Person2.get_count() == 4


def test_get_count_none(use_graph):
    assert Person2.get_count() == 0


def test_merge_empty_df():
    df = pd.DataFrame()

    result = PracticeNode.merge_df(df)

    assert len(result) == 0
    assert isinstance(result, pd.Series)


class SampleEnum(Enum):
    VALUE1 = "value1"
    VALUE2 = "value2"
    VALUE3 = "value3"


class AugmentedPerson(BaseNode):
    __primaryproperty__: ClassVar[GQLIdentifier] = "name"
    __primarylabel__: ClassVar[GQLIdentifier] = "AugmentedPerson"

    name: str
    optional_enum: Optional[SampleEnum] = Field(
        default_factory=lambda: SampleEnum.VALUE1
    )

    @field_serializer("optional_enum")
    def serialize_enum(self, value: Optional[SampleEnum]) -> Optional[str]:
        return value.value if value is not None else None

    @related_nodes
    def followers(self):
        return "MATCH (#ThisNode)<-[:AUGMENTED_PERSON_FOLLOWS]-(o) RETURN o"

    @related_property
    def follower_count(self):
        return "MATCH (#ThisNode)<-[:AUGMENTED_PERSON_FOLLOWS]-(o) RETURN COUNT(DISTINCT o)"

    @property
    @related_property
    def follower_names(self):
        return "MATCH (#ThisNode)<-[:AUGMENTED_PERSON_FOLLOWS]-(o) RETURN COLLECT(DISTINCT o.name)"


class AugmentedPersonRelationship(BaseRelationship):
    __relationshiptype__: ClassVar[str] = "AUGMENTED_PERSON_FOLLOWS"

    source: AugmentedPerson
    target: AugmentedPerson

    follow_tag: Optional[str] = None


def test_get_related_node_methods():
    assert set(AugmentedPerson.get_related_node_methods().keys()) == {
        "followers",
    }


def test_get_related_prop_methods():
    # note that only decorated methods, not properties are returned
    assert set(AugmentedPerson.get_related_property_methods().keys()) == {
        "follower_count",
        "get_count",
    }


def test_node_schema():
    schema = AugmentedPerson.neontology_schema()

    assert schema.properties[0].name == "name"
    assert schema.properties[0].required is True
    assert schema.outgoing_relationships[0].name == "AUGMENTED_PERSON_FOLLOWS"


def test_node_schema_json():

    schema_json = AugmentedPerson.neontology_schema().model_dump_json()

    schema_dict = json.loads(schema_json)

    assert schema_dict["properties"][1]["type_annotation"]["core_type"] == "SampleEnum"

    assert schema_dict["properties"][1]["type_annotation"]["enum_values"] == [
        "value1",
        "value2",
        "value3",
    ]


def test_node_schema_md():
    schema = AugmentedPerson.neontology_schema()

    schema_md = schema.md_node_table()

    assert "| Property Name | Type | Required |" in schema_md
    assert "| name | str | True |" in schema_md


def test_rels_schema_md():
    schema = AugmentedPerson.neontology_schema()

    schema_md = schema.md_rel_tables(heading_level=4)

    assert "#### AUGMENTED_PERSON_FOLLOWS" in schema_md
    assert "| follow_tag | Optional[str] | False |" in schema_md


def test_related_nodes(use_graph):
    alice = AugmentedPerson(name="Alice")
    alice.merge()

    bob = AugmentedPerson(name="Bob")
    bob.merge()

    follows = AugmentedPersonRelationship(
        source=alice, target=bob, follow_tag="test-tag"
    )
    follows.merge()

    follows2 = AugmentedPersonRelationship(
        source=bob, target=alice, follow_tag="second-tag"
    )
    follows2.merge()

    alice_rels = alice.get_related()

    related_nodes = [x for x in alice_rels.nodes if x.get_pp() != alice.get_pp()]

    assert len(related_nodes) == 1
    assert related_nodes[0].name == "Bob"

    bobs_followers = bob.get_related(
        relationship_types=["AUGMENTED_PERSON_FOLLOWS"],
        incoming=True,
        outgoing=False,
        relationship_properties={"follow_tag": "test-tag"},
    )

    assert len(bobs_followers.nodes) == 2  # this will include bob himself
    assert bobs_followers.nodes[0].name == "Alice"

    bobs_rels = bob.get_related(incoming=True, distinct=True)

    assert len(bobs_rels.nodes) == 2
    assert bobs_rels.nodes[0].name == "Alice"

    assert len(bobs_rels.relationships) == 2


def test_related_nodes_unmerged(use_graph):
    alice = AugmentedPerson(name="Alice")

    alice_rels = alice.get_related()

    assert len(alice_rels.nodes) == 0


def test_related_nodes_no_rels(use_graph):
    alice = AugmentedPerson(name="Alice")
    alice.merge()

    bob = AugmentedPerson(name="Bob")
    bob.merge()

    alice_rels = alice.get_related()

    assert len(alice_rels.nodes) == 0


def test_retrieve_property(use_graph):
    alice = AugmentedPerson(name="Alice")
    alice.merge()

    bob = AugmentedPerson(name="Bob")
    bob.merge()

    follows = AugmentedPersonRelationship(
        source=alice, target=bob, follow_tag="test-tag"
    )
    follows.merge()

    assert bob.follower_count() == 1
    assert bob.follower_names == ["Alice"]


def test_retrieve_property_none(use_graph):
    alice = AugmentedPerson(name="Alice")
    alice.merge()

    bob = AugmentedPerson(name="Bob")
    bob.merge()

    assert bob.follower_count() == 0
    assert not bob.follower_names


def test_retrieve_nodes_none(use_graph):
    alice = AugmentedPerson(name="Alice")
    alice.merge()

    bob = AugmentedPerson(name="Bob")
    bob.merge()

    followers = bob.followers()

    assert len(followers) == 0


class ComplexPerson(BaseNode):
    __primaryproperty__: ClassVar[str] = "identifier"
    __primarylabel__: ClassVar[str] = (
        "PersonLabel1"  # optionally specify the label to use
    )

    name: str = Field(default_factory=uuid4)
    age: int
    favorite_colors: list = ["red", "green", "blue"]
    favorite_numbers: list = [1, 2, 3]
    extra_str1: UUID = Field(default_factory=uuid4)
    extra_str2: UUID = Field(default_factory=uuid4)

    identifier: Optional[str] = Field(default=None, validate_default=True)

    @field_validator("identifier")
    def set_identifier(cls, v, values):
        if v is None:
            v = f"{values.data['name']}_{values.data['age']}"

        return v

    @field_serializer("extra_str1", "extra_str2")
    def serialize_to_str(self, v: UUID):
        return str(v)


def test_create_mass_nodes(use_graph, benchmark):
    people_records = [{"age": x, "name": uuid4().hex} for x in range(1000)]

    people_df = pd.DataFrame.from_records(people_records)

    benchmark(ComplexPerson.merge_df, people_df)

    assert Person.get_count() == 1000


class UserWithAliases(BaseNode):
    __primaryproperty__: ClassVar[str] = "userName"
    __primarylabel__: ClassVar[str] = "User"
    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
        populate_by_name=True,  # allow population by name and alias
    )
    user_name: Annotated[str, Field(alias="userName")]
    some_other_property: Annotated[Optional[str], Field(None, alias="otherProperty")]


def test_aliased_properties(use_graph):
    user1: UserWithAliases = UserWithAliases(userName="User1")
    user2: UserWithAliases = UserWithAliases(
        user_name="User2", some_other_property="alpha"
    )
    user3: UserWithAliases = UserWithAliases(userName="User3", otherProperty="beta")
    assert user1.user_name == "User1"
    assert user3.some_other_property == "beta"

    user1.merge()
    user2.merge()
    user3.merge()

    cypher = """
    MATCH (n:User)
    RETURN n
    ORDER BY n.userName ASC
    """

    result: NeontologyResult = use_graph.evaluate_query(cypher)
    assert result.nodes[0].user_name == "User1"
    assert hasattr(result.nodes[0], "userName") == False
    assert result.records_raw[0][0]["userName"] == "User1"
    assert result.nodes[0].some_other_property is None
    assert result.records_raw[0][0]["otherProperty"] is None

    assert result.nodes[1].user_name == "User2"
    assert result.nodes[1].some_other_property == "alpha"
    assert result.records_raw[1][0]["otherProperty"] == "alpha"

    assert result.nodes[2].user_name == "User3"
    assert result.records_raw[2][0]["otherProperty"] == "beta"



from __future__ import annotations

import enum
from logging import getLogger
from typing import Any, Optional, Union, get_args, get_origin

from jinja2 import Template
from pydantic import BaseModel, field_serializer

logger = getLogger(__name__)


class NeontologyAnnotationData(BaseModel):
    representation: str
    core_type: Any
    optional: bool = False
    union: bool = False
    enum_values: Optional[list] = None

    @field_serializer("core_type")
    def serialize_core_type(self, value: Any) -> str:
        """Serialize the core type to a string representation."""
        if isinstance(value, type):
            return value.__name__
        elif isinstance(value, enum.Enum):
            return value.__name__
        else:
            return str(value)


class SchemaProperty(BaseModel):
    name: str
    type_annotation: NeontologyAnnotationData
    required: bool


class RelationshipSchema(BaseModel):
    name: str
    relationship_type: str
    source_labels: list[str]
    target_labels: list[str]

    properties: list[SchemaProperty]


class NodeSchema(BaseModel):
    label: str
    title: str
    secondary_labels: list[str]
    properties: list[SchemaProperty]
    outgoing_relationships: list[RelationshipSchema] = []

    def md_node_table(self) -> str:
        """Take a node schema and produce markdown ontology documentation."""
        schema_template_raw = """
| Property Name | Type | Required |
| ------------- | ---- | -------- |
{% for field in model_schema.properties -%}
| {{field.name}} | {{field.type_annotation.representation}} | {{field.required}} |
{% endfor %}
"""

        schema_template = Template(schema_template_raw)

        return schema_template.render(model_schema=self).strip()

    def md_rel_tables(self, heading_level: int = 3) -> str:
        """Take a node schema and produce markdown ontology documentation."""
        schema_template_raw = """
{% if model_schema.outgoing_relationships %}
{% for outgoing_rel in model_schema.outgoing_relationships -%}
{{"#"*heading_level}} {{ outgoing_rel.relationship_type }}

Target Label(s): {{ outgoing_rel.target_labels |join(', ') }}
{% if outgoing_rel.properties %}
| Property Name | Type | Required |
| ------------- | ---- | -------- |
{% for rel_prop in outgoing_rel.properties -%}
| {{rel_prop.name}} | {{rel_prop.type_annotation.representation}} | {{rel_prop.required}} |
{% endfor %}
{% endif %}

{% endfor %}
{% endif %}
"""

        schema_template = Template(schema_template_raw)

        return schema_template.render(model_schema=self, heading_level=heading_level).strip()


def extract_type_mapping(annotation: Any, show_optional: bool = True) -> NeontologyAnnotationData:
    """Extract type information from a type annotation."""
    print(annotation)

    if isinstance(annotation, type):
        if issubclass(annotation, enum.Enum):
            enum_values = [e.value for e in annotation]

            return NeontologyAnnotationData(
                representation="Enum",
                core_type=annotation,
                enum_values=enum_values,
            )

        else:
            # we have a plain type, just return the name

            return NeontologyAnnotationData(
                representation=str(annotation.__name__),
                core_type=annotation,
            )

    elif get_origin(annotation) == Union:
        # We can only support union's of a single type plus none (i.e. Optional)
        if len(get_args(annotation)) == 2 and type(None) in get_args(annotation):
            # we need to extract the optional type
            for entry in get_args(annotation):
                if isinstance(entry, type(None)):
                    pass
                else:
                    # we do this recursively in case the next layer down is something like list[int]
                    if show_optional is True:
                        representation = f"Optional[{extract_type_mapping(entry).representation}]"
                        extracted_entry = extract_type_mapping(entry)
                        core_type = extracted_entry.core_type
                        print(f"Extracted optional type: {representation}, core type: {core_type}")
                        return NeontologyAnnotationData(
                            optional=True,
                            representation=representation,
                            core_type=core_type,
                            enum_values=extracted_entry.enum_values,
                        )
                    else:
                        return extract_type_mapping(entry)
                raise TypeError(f"Unsupported type annotation: {annotation}")
        else:
            raise TypeError(f"Unsupported union type: {annotation}")

    elif get_origin(annotation) is list:
        if len(get_args(annotation)) == 1:
            try:
                # field type is something like typing.list[str]
                representation = str(annotation).split(".")[1]

            except IndexError:
                # some Python versions support just list[x]
                representation = str(annotation).title()
            return NeontologyAnnotationData(
                representation=representation,
                core_type=annotation,
            )
        else:
            raise TypeError(f"Cannot have lists of multiple types: {annotation}")

    logger.warning(f"Complex type annotation: {annotation}")
    return NeontologyAnnotationData(representation=str(annotation.__name__), core_type=annotation)
